package Globus::GRAM::JobManager::Helper;

@Globus::GRAM::JobManager::Helper::ISA = qw(Globus::GRAM::JobManager);

use Globus::GRAM::JobState;
use Globus::GRAM::JobManager;
use Globus::Core::Paths;
use Globus::GRAM::JobDescription;

#use strict;
use IO::File;
use Fcntl ":flock";
use Fcntl;

use POSIX qw(:sys_wait_h);
use POSIX qw(sysconf _SC_OPEN_MAX setsid fstat);

#
# Don't use GPG for now because of problem with update_proxy() sometimes not being called
#
my $g_useGPG = 0;

# Loading constants
my $FINAL_DONE_MIN_LOAD = 10;
my $FINAL_DONE_LOAD_RANGE = 10;
my $PARENT_PAUSE_MIN_LOAD = 10;
my $PARENT_PAUSE_LOAD_RANGE = 10;
my $MAX_PAUSE_PERIOD = 2700;
my $MIN_MEM_FREE = 0.2;
my $MAX_IMPORTING_CYCLE_LEN = 60;
my $MAX_SUBMIT_CYCLE_LEN = 60;
my $MIN_SUBMIT_AGE = 30;
my $UPTIME_REFRESH = 15;
my $MAX_GRANDCHILD_LOAD = 20;
my $MAX_GRANDCHILD_SFACT = 2;
my $MAX_POLL_AGE = 55;
my $MAX_FIN_TIME = 86400;
my $MAX_POLL_PROC_AGE = 1800;
my $POLL_PROC_LIFE = 180;

# Misc constants
my $LOCK_CHECK = 25;
my $INCOMING_DIR_CHECK = 100;
my $LRMS_QUERY_DELAY = 15;
my $MAX_ACCT_WAIT = 300;

sub new
{
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@_);

    bless $self, $class;
    return $self;
}

sub helper_get_url_and_tags
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";

    my ($tag,$url,$hash_ref) = @_;

    my $cache_url;

    my $qual = "";
    $qual = "-t $tag" if $tag ne "";

    local(*CACHEPRG);
    open(CACHEPRG,$cache_pgm." -list ".$qual." $url |");
    while(<CACHEPRG>)
    {
        chomp(my $line=$_);
        if ($line =~ /^\s*URL:\s*(\S+)/)
        {
            $cache_url = $1;
        }
        if (defined $cache_url && $line =~ /^\s*Tag:\s*(\S+)/)
        {
            my $tag = $1;
            if (defined $hash_ref->{$cache_url})
            {
                $hash_ref->{$cache_url} .= " ".$tag;
            }
            else
            {
                $hash_ref->{$cache_url} = $tag;
            }
            $cache_url = undef;
        }
    }
    close(CACHEPRG);
}

sub helper_get_cache_export_dir
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};
    my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";
    my ($url_ref) = @_;

    return undef unless defined $tag;

    my $cache_export_url;
    $cache_export_url = "$tag/gk_cache_export_dir";

#    local(*CACHEPRG);
#    open(CACHEPRG,"$cache_pgm -list -t $tag |");
#    while(<CACHEPRG>)
#    {
#        chomp(my $line=$_);
#        if ($line =~ /^\s*URL:\s*(\S+)/)
#        {
#             my $url = $1;
#             $cache_export_url = $url if $url =~ m|^$tag/gk_cache_export_dir|;
#        }
#    }
#    close(CACHEPRG);

    if (defined $cache_export_url)
    {
        $$url_ref = $cache_export_url if ref $url_ref;

        chomp(my $cache_export_ptr_file = `$cache_pgm -query -t $tag $cache_export_url`);
        if($cache_export_ptr_file ne "")
        {
            local(*CACHEPTR);
            open(CACHEPTR,"< ".$cache_export_ptr_file);
            my $cache_export_dir;
            while(<CACHEPTR>)
            {
              chomp($cache_export_dir = $_);
            }
            close(CACHEPTR);
            if (defined $cache_export_dir)
            {
                return $cache_export_dir;
            }
        }
    }

    return undef;
}

sub helper_cache_export
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};
    my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";

    my $cache_export_dir = $self->helper_get_cache_export_dir();
    return if !defined $cache_export_dir || $cache_export_dir eq "";

    my (@stage_in_list,@stage_in_shared_list);
    $self->helper_get_from_tmp_file("stage_in",\@stage_in_list);
    $self->helper_get_from_tmp_file("stage_in_shared",\@stage_in_shared_list);

    my (@extra_cache_entries,@local_stage_in_all);
    foreach my $pair (@stage_in_shared_list)
    {
        my ($remote,$local) = split('\s+',$pair);
        push(@extra_cache_entries,$remote);
        push(@local_stage_in_all,$local);
    }
    foreach my $pair (@stage_in_list)
    {
        my ($remote,$local) = split('\s+',$pair);
        push(@local_stage_in_all,$local);
    }
    if ($description->executable() =~ m|^[a-zA-Z]+://|)
    {
        push(@extra_cache_entries,$description->executable());
    }
    if ($description->stdin() =~ m|^[a-zA-Z]+://|)
    {
        push(@extra_cache_entries,$description->stdin());
    }

    my %cache_url_tags;
#
# Should scan the gass cache for every entry associated to our tag
# However the -list option for the gass_cache_util does not
# scale well with cache size so we avoid the scan
#
#    $self->helper_get_url_and_tags($tag,"",\%cache_url_tags);

#
# Instead limit the check to the following urls, listed explicitly
#
    my @tmp_array = split("/",$tag);
    my @tmp_array2 = split(":",$tmp_array[2]);
    my $xgass = "x-gass-cache://".$tmp_array2[0]."/".$tmp_array[3].".".$tmp_array[4]."/";
    foreach my $url ($xgass."cache-info", $tag."x509_user_proxy", $tag."/gk_cache_export_dir",
                     $xgass."dev/stdout", $xgass."dev/stderr", @extra_cache_entries)
    {
        chomp(my $file = `$cache_pgm -query -t $tag $url`);

        if ($file ne '')
        {
             $cache_url_tags{$url} = $tag;
        }
    }

    my $export_fn = $cache_export_dir."/export.txt";

    my %url_seen;
    my $nexports = 0;

    local(*CACHEEXPORT);
    open(CACHEEXPORT,"> ".$export_fn);
    foreach my $cache_url (keys %cache_url_tags)
    {
        my @tags =  split('\s+',$cache_url_tags{$cache_url});
        foreach my $this_tag (@tags)
        {
            local(*CACHEPRG);
            open(CACHEPRG,$cache_pgm." -query -t $this_tag $cache_url |");
            my $file;
            while(<CACHEPRG>)
            {
                chomp($file = $_);
            }
            close(CACHEPRG);
            if (defined $file && $file ne "")
            {
                if (!defined  $url_seen{$cache_url})
                {
                    $url_seen{$cache_url} = 1;
                    $nexports++;
                    symlink($file,$cache_export_dir."/export.".$nexports);
                }
                print CACHEEXPORT $this_tag." ".$cache_url." ".$file." "."export.".$nexports."\n";
            }
        }
    }
    close(CACHEEXPORT);

    local(*FILELIST);
    open(FILELIST,"> ".$cache_export_dir."/stage_in.txt");
    foreach my $line (@local_stage_in_all)
    {
        print FILELIST $line."\n";
    }
    close(FILELIST);

    my @tmp_list=();
    $self->helper_get_from_tmp_file("stage_out",\@tmp_list);

    open(FILELIST,"> ".$cache_export_dir."/stage_out.txt");
    foreach ($description->file_stage_out(),@tmp_list)
    {
        next unless defined $_;
        if (ref $_)
        {
            my ($local, $remote) = ($_->[0], $_->[1]);
            print FILELIST $local."\n";
        }
        else
        {
            print FILELIST $_."\n";
        }
    }
    close(FILELIST);

    @tmp_list=();
    $self->helper_get_from_tmp_file("file_cleanup",\@tmp_list);
    open(FILELIST,"> ".$cache_export_dir."/file_cleanup.txt");
    foreach my $clean ($description->file_cleanup(),@tmp_list)
    {
        next unless defined $clean;
        next if ref($clean);

        print FILELIST $clean."\n";
    }
    close(FILELIST);

    open(FILELIST,"> ".$cache_export_dir."/stdstreams.txt");
    if ($description->stdout() ne '')
    {
        print FILELIST $description->stdout()."\n";
    }
    else
    {
        print FILELIST "\n";
    }
    if ($description->stderr() ne '')
    {
        print FILELIST $description->stderr()."\n";
    }
    else
    {
        print FILELIST "\n";
    }
    close(FILELIST);

    $self->cd_fork_and_exec_cmd($cache_export_dir,0,0,"/bin/tar","cf","cache_export_dir.tar","export.txt");

    my %exclude;
    foreach(".","..","export.txt","cache_export_dir.tar")
    {
        $exclude{$_} = 1;
    }

    local(*CACHEDIR);
    opendir(CACHEDIR,$cache_export_dir);
    foreach my $dir_entry (readdir(CACHEDIR))
    {
        next if defined $exclude{$dir_entry};
        $self->cd_fork_and_exec_cmd($cache_export_dir,0,0,"/bin/tar","rhf","cache_export_dir.tar",$dir_entry);
    }
    closedir(CACHEDIR);
}

sub valid_lockfile
{
    my $self = shift;
    my ($dir,$fn,$my_host,$host,$pid) = @_;

    return 0 if ! -e $dir."/".$fn;
    return 1 if ($host ne "" && $my_host ne $host);

    if ($host ne "" && kill(0,$pid) == 1)
    {
        return 1;
    }
    else
    {
        unlink($dir."/".$fn);
        return 0;
    }
}

sub lock_ptr_to_handle
{
    my $self = shift;
    my ($dir,$lock_ptr) = @_;

    my $inode = $self->test_lock_ptr($lock_ptr);
    if ($inode != 0)
    {
        my ($lock_fname,$hold_fname) = split(":",$lock_ptr);
        local(*LF);
        if (open(LF,"< ".$lock_fname))
        {
            my $info = <LF>;
            chomp($info);
            ($lock_id,$hold_fn) = split(":",$info);
            close(LF);
            if ($hold_fn eq $hold_fname)
            {
                return $dir.":".$lock_id.":".$lock_fname.":".$inode;
            }
        }
    }

    return undef;
}

sub l_lock_lock_handle
{
    my $self = shift;
    my ($lock_handle) = @_;

    return undef if !defined $lock_handle;

    my ($dir,$lid,$lock_fn,$inode) = split(":",$lock_handle);

    my $hl = $self->get_lock($dir,"lock-$lid-$inode",undef);

    my $lock_ptr = $self->lock_handle_deref($lock_handle);
    if (!defined $lock_ptr)
    {
        $self->l_release_handle_lock($hl);
        return undef;
    }

    return $hl;
}

sub l_release_handle_lock
{
    my $self = shift;
    my ($hl) = @_;

    my $lock_ptr = $self->lock_handle_deref($hl);
    return if !defined $lock_ptr;

    my ($lock_fname,$hold_fname) = split(":",$lock_ptr);

    unlink($lock_fname);
    unlink($hold_fname);

    return;
}

sub lock_handle_deref
{
    my $self = shift;
    my ($lock_handle) = @_;

    return undef if !defined $lock_handle;

    my ($dir,$lid,$lock_fn,$inode) = split(":",$lock_handle);
    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = CORE::stat($lock_fn);

    return undef if (!defined $ino || $ino ne $inode);

    my ($hold_fn,$lock_id);
    local(*LF);
    if (open(LF,"< ".$lock_fn))
    {
        my $info = <LF>;
        chomp($info);
        ($lock_id,$hold_fn) = split(":",$info);
        close(LF);
    }
    else
    {
        return undef;
    }

    return undef if $lock_id ne $lid;

    my ($lock_fname,$hold_fname) = ($lock_fn,$hold_fn);
    return $lock_fname.":".$hold_fname;
}

sub test_lock_ptr
{
    my $self = shift;
    my ($lock_ptr) = @_;

    return 0 if !defined $lock_ptr;

    my ($dev1,$dev2,$ino1,$ino2);
    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks);

    my ($lock_fname,$hold_fname) = split(":",$lock_ptr);
    ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = CORE::stat($hold_fname);
    $dev1 = $dev; $ino1 = $ino;

    ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = CORE::stat($lock_fname);
    $dev2 = $dev; $ino2 = $ino;

    return 0 if (!defined $dev1 || !defined $dev2 || !defined $ino1 || !defined $ino2 || $dev1 != $dev2 || $ino1 != $ino2);

    return $ino1;
}

sub getpid_lock
{
    my $self = shift;
    my ($lock_handle) = @_;

    my $hl = $self->l_lock_lock_handle($lock_handle);
    my $lock_ptr = $self->lock_handle_deref($lock_handle);
    $self->l_release_handle_lock($hl);

    return undef if !defined $lock_ptr;

    chomp(my $hn = `hostname`);
    $hn =~ s/\..*//;

    my ($lock_fname,$hold_fname) = split(":",$lock_ptr);

    my @tmp = split("/",$hold_fname);
    my $hold = pop(@tmp);

    if ($hold =~ m|^(\S+)\.hold\.(\S*)\.(\d+)|)
    {
        my ($lock_name,$host,$pid) = ($1,$2,$3);
        return $pid if ($host eq $hn);
    }

    return undef;
}

sub chown_lock
{
    my $self = shift;
    my ($lock_handle) = @_;

    my $hl = $self->l_lock_lock_handle($lock_handle);
    my $lock_ptr = $self->lock_handle_deref($lock_handle);

    if (!$self->test_lock_ptr($lock_ptr))
    {
        $self->l_release_handle_lock($hl);
        return undef;
    }

    chomp(my $hn = `hostname`);
    $hn =~ s/\..*//;

    my ($lock_fname,$hold_fname) = split(":",$lock_ptr);
    my $new_hold_fname = $hold_fname;

    my @tmp = split("/",$hold_fname);
    my $hold = pop(@tmp);
    my $base = join("/",@tmp);

    if ($hold =~ m|^(\S+)\.hold\.(\S*)\.(\d+)|)
    {
        my ($lock_name,$host,$pid) = ($1,$2,$3);
        $new_hold_fname = $base."/".$lock_name.".hold.".$hn.".".$$;
    }

    my $ret = undef;
    local(*LF);
    if (open(LF,"+< $lock_fname"))
    {
        my $info = <LF>;
        chomp($info);
        my ($lid,$hold_fn) = split(":",$info);
        seek(LF, 0, 0);

        print LF $lid.":".$new_hold_fname."\n";
        close(LF);
        rename $hold_fname, $new_hold_fname;
        $ret = 1;
    }

    $self->l_release_handle_lock($hl);

    return $ret;
}

sub release_lock
{
    my $self = shift;
    my ($lock_handle) = @_;

    my $hl = $self->l_lock_lock_handle($lock_handle);
    my $lock_ptr = $self->lock_handle_deref($lock_handle);

    if (!$self->test_lock_ptr($lock_ptr))
    {
        $self->l_release_handle_lock($hl);
        return undef;
    }

    my ($lock_fname,$hold_fname) = split(":",$lock_ptr);

    unlink($lock_fname);
    unlink($hold_fname);

    $self->l_release_handle_lock($hl);

    #open(TMP,">> /tmp/t1233.t");
    #print TMP time()." $$ released a lock handle=$lock_handle\n";
    #close(TMP);

    return;
}

sub l_scan_dir
{
    my $self = shift;
    my ($dir,$lock_name,$hn,$lock_devino) = @_;

    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks);

    my $valid_lock = -1;

    local(*DIR);
    if (opendir(DIR,$dir))
    {
        $valid_lock = 0;
        foreach my $entry (readdir(DIR))
        {
            if ($entry =~ /\.tmp\.(\S*)\.(\d+)$/)
            {
                my ($host,$pid) = ($1,$2);
                $self->valid_lockfile($dir,$entry,$hn,$host,$pid);
            }

            if ($entry =~ /^$lock_name\.hold\.(\S*)\.(\d+)$/)
            {
                my ($host,$pid) = ($1,$2);
                if ($self->valid_lockfile($dir,$entry,$hn,$host,$pid))
                {
                    my $test_tmp = $lock_name.".tmp.".$host.".".$pid;
                    if ( ! $self->valid_lockfile($dir,$test_tmp,$hn,$host,$pid))
                    {
                        ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = CORE::stat($dir."/".$entry);
                        if (defined $dev && defined $ino)
                        {
                            my $test_devino = $dev.":".$ino;
                            $valid_lock++ if (defined $lock_devino && $test_devino eq $lock_devino);
                        }
                    }
                }
            }
        }
        closedir(DIR);
    }

    return $valid_lock;
}
    
sub get_lock
{   
    my $self = shift;
    my ($dir,$lock_name,$retry_count) = @_;

    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks);

    chomp(my $hn = `hostname`);
    $hn =~ s/\..*//;

    mkdir $dir, 0700 if ! -d $dir;

    my $hold_fname = $dir."/".$lock_name.".hold.".$hn.".".$$;
    my $tmp_fname = $dir."/".$lock_name.".tmp.".$hn.".".$$;
    my $lock_fname = $dir."/".$lock_name.".locked"; 

    my $lock_ptr = $lock_fname.":".$hold_fname;
    my $handle;
#
# Return undef if already locked
#
    if ($self->test_lock_ptr($lock_ptr))
    {
        return undef;
    }

    my $start = time();
    my $got_lock = 0;

    while(!$got_lock)
    {
        $self->l_scan_dir($dir,$lock_name,$hn,undef) if (rand($LOCK_CHECK)<1);

        local(*LF);
        open(LF, ">> $lock_fname");
        close(LF);

        ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = CORE::stat($lock_fname);
        next if !defined $nlink || $nlink > 2;

        link $lock_fname, $tmp_fname;
        ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = CORE::stat($tmp_fname);

        if (!defined $nlink || $nlink > 3)
        {
            unlink($tmp_fname) if defined $nlink;
            next;
        }

        my $lock_devino = $dev.":".$ino;

        my $valid_lock = $self->l_scan_dir($dir,$lock_name,$hn,$lock_devino);

        if ($valid_lock >= 0)
        {
            link $lock_fname, $hold_fname;
            ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = CORE::stat($lock_fname);

            my $lock2_devino;
            $lock2_devino = $dev.":".$ino if defined $dev && defined $ino;

            my $inconsistent = 0;
            if (!defined $lock2_devino || $lock2_devino ne $lock_devino || ! -e $hold_fname )
            {
                $inconsistent++;
                $valid_lock = 0;
            }

            if (!$inconsistent && !$valid_lock && $nlink==3 && open(LF, "> $lock_fname"))
            {
                my $lock_id = $self->l_get_rand();
                print LF $lock_id.":".$hold_fname."\n";
                close(LF);
                $handle = $self->lock_ptr_to_handle($dir,$lock_ptr);
                $got_lock++ if defined $handle;
            }

            unlink($hold_fname) if !$got_lock;
            unlink($tmp_fname);

            if (!$got_lock)
            {
                if (defined $retry_count && $retry_count>=0 && $valid_lock && $retry_count<=(time() - $start))
                {
                    return undef;
                }
            }
        }
        else
        {
            unlink($tmp_fname);
            return undef;
        }
    }
    continue
    {
         my $wait_time = 0.25;
         $wait_time = -1*$retry_count if defined $retry_count && $retry_count<0;

         $self->l_my_sleep($wait_time + $wait_time*(rand(2) - 1.0)) if !$got_lock;
    }

    #open(TMP,">> /tmp/t1233.t");
    #print TMP time()." $$ get a lock ($lock_name) handle=$handle\n";
    #close(TMP);

    return $handle;
}

sub helper_cache_import
{
    my $self = shift;
    my ($internal_id,$batch_id,$submit_time,$state_ref,$batchout,$batcherr) = @_;
    my $description = $self->{JobDescription};
    my $uniq_id = $description->uniq_id();
    my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};
    my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";

    my %running;
    $self->l_list_read("jm-running.list",\%running);

    if ($$state_ref == Globus::GRAM::JobState::ACTIVE)
    {

        if (!exists $running{$uniq_id})
        {
            my $lock_handle = $self->get_lock(".lcgjm","jm-running.list",undef);
            $self->l_list_read_add("jm-running.list",$uniq_id,undef,"RUNNING",\%running);
            $self->release_lock($lock_handle);
        }
    }

    my $import_ret = $self->l_helper_cache_import($$state_ref,$batchout,$batcherr);
    if ($import_ret <= 0)
    {
#        $$state_ref = (exists $running{$uniq_id}) ? Globus::GRAM::JobState::ACTIVE : Globus::GRAM::JobState::PENDING;
        $$state_ref = Globus::GRAM::JobState::ACTIVE;
    }
    elsif ($import_ret == 3)
    {
        $$state_ref = Globus::GRAM::JobState::FAILED;
    }
    elsif ($import_ret == 1)
    {
        if (defined $batch_id)
        {
            my $running_gm = $self->l_is_gm();

            if ($running_gm == $$)
            {
                my @loadavg;
                $self->get_loadavg(\@loadavg);
                my $prob = ($loadavg[0] - $FINAL_DONE_MIN_LOAD)/$FINAL_DONE_LOAD_RANGE;
                $prob = (($prob+0.5)*($prob+0.5)-0.25)/2 if $prob>0;

                if (rand(1) < $prob)
                {
#                    $$state_ref = (exists $running{$uniq_id}) ? Globus::GRAM::JobState::ACTIVE : Globus::GRAM::JobState::PENDING;
                    $$state_ref = Globus::GRAM::JobState::ACTIVE;
                    return -2;
                }
            }

            my @acct_info;
            $self->l_tidy_queues();

            my $acct_url = "$tag/job-acct.info";
            my $acct_file_name = `$cache_pgm -query -t $tag $acct_url`;
            chomp($acct_file_name);
 
            if ($acct_file_name ne '' && open(ACCT,"< ".$acct_file_name))
            {
                while(<ACCT>)
                {
                    chomp(my $line=$_);
                    $line =~ s|\\|\\\\|g;
                    push(@acct_info,$line);
                }
                close(ACCT);
            }

            my $orig_pid = ($self->l_is_gm() != $$) ? getppid() : $$;
            $self->write_to_gacct($orig_pid,"Job $internal_id (ID $batch_id) has finished");

            if (scalar(@acct_info)>0)
            {
                $self->write_to_gacct($orig_pid,"summary:");
                foreach my $line (@acct_info)
                {
                    $self->write_to_gacct($orig_pid,$line);
                }
                $self->write_to_gacct($orig_pid,"-- end of summary");
            }
        }
        else
        {
            my %finished_list;
            $self->l_list_read("jm-finished.list",\%finished_list);
            if (!exists $finished_list{$tag} && $self->l_exists_gram_state($uniq_id))
            {
                $self->l_tidy_queues();
            }
        }
    }
    elsif ($import_ret == 2)
    {
        if ($$state_ref == Globus::GRAM::JobState::FAILED)
        {
            $self->l_import_cache($$state_ref,$batchout,$batcherr);
        }
    }

#    if (defined $batch_id)
    if (0)
    {
	#
	# the l_delay_parent() method is suspected of occasionally
	# leaving grid_monitor or job manager processes suspended;
	# since this code was foreseen to be used only in absence
	# of a grid_monitor, David Smith suggested commenting it out
	#

        my @loadavg;
        $self->get_loadavg(\@loadavg);

        if ($loadavg[0] > $PARENT_PAUSE_MIN_LOAD)
        {
            my $running_gm = $self->l_is_gm();

            if ($running_gm != $$)
            {
                $self->l_delay_parent(getppid(),$loadavg[0]);
            }
        }
    }

    return $import_ret;
}

#
# Cache import: returns:
#
#   -1 = Job added to import queue. An importing child already exists
#    0 = Job added to import queue. Child process has been started to import the cache.
#    1 = No cache associated with the job
#    2 = Status is not DONE
#    3 = fork() error or problem writing to import list
#    4 = job has inconsistent entries in submit_list or dq_list
#
sub l_helper_cache_import
{
    my $self = shift;
    my ($status,$batchout,$batcherr) = @_;
    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};
    my $job_id = $description->jobid();
    my %import_list;
    my $child;

    do
    {
        $child = waitpid(-1,WNOHANG);
    } until $child <= 0;

    return 2 if $status != Globus::GRAM::JobState::DONE;

    my $cache_export_dir = $self->helper_get_cache_export_dir();
    $self->l_list_read("jm-import.list",\%import_list);
    if (!exists $import_list{$job_id} && !exists $import_list{$job_id."-acct"})
    {
        return 1 if ( !defined $cache_export_dir || $cache_export_dir eq "" || ! -d $cache_export_dir);
    }

    my ($queue_epoch,$jm_type,$iid,$iuqid) = split(":",$job_id);
    my $list_name = "jm-".$jm_type."-submit.list";
    my $dq_name = "jm-".$jm_type."-dequeue.list";
    my $list_lock_handle =  $self->get_lock(".lcgjm",$list_name,undef);

    {
        my (%submit_list,%dq_list);
        $self->l_list_read($list_name,\%submit_list);
        $self->l_list_read($dq_name,\%dq_list);

        if (!exists $submit_list{$job_id} || exists $dq_list{$job_id})
        {
            $self->release_lock($list_lock_handle);
            return 4;
        }
    }

    my $implist_lock_handle = $self->get_lock(".lcgjm","jm-import.list",undef);
    $self->l_list_read("jm-import.list",\%import_list);

    if (!exists $import_list{$job_id} && !exists $import_list{$job_id."-acct"})
    {
        $description->add('importstatus', $status);
        $description->add('importbatchout', $batchout);
        $description->add('importbatcherr', $batcherr);
        $description->add('importstarttime', time());

        my $err = 0;
        $err = 1 if ( !defined $cache_export_dir || $cache_export_dir eq "" || ! -d $cache_export_dir);
        $err = 3 if $err==0 && !$self->l_list_read_add("jm-import.list",$job_id,undef,undef,\%import_list);
        if ($err==0 && !$self->l_list_read_add("jm-import.list",$job_id."-acct",undef,undef,\%import_list))
        {
            $self->l_list_read_remove("jm-import.list",$job_id,\%import_list);
            $err = 3;
        }

        if ( $err )
        {
            $self->release_lock($list_lock_handle);
            $self->release_lock($implist_lock_handle);
            return $err;
        }
    }

    $self->release_lock($list_lock_handle);

    my $external_lock_handle = $self->get_lock(".lcgjm","jm-import.proc",0);
    $self->release_lock($implist_lock_handle);
    $implist_lock_handle = undef;

    if (!defined $external_lock_handle)
    {
        return -1;
    }

    $child = fork();
    if (!defined $child)
    {
        $self->release_lock($external_lock_handle);
        return 3;
    }

    if ($child == 0)
    {
        $self->l_daemonise();

        $self->chown_lock($external_lock_handle);
    }
    else
    {
        my $pid;
        do
        {
            $pid = $self->getpid_lock($external_lock_handle);
               
        } while(((defined $pid && $pid != $child)) && ($self->l_my_sleep(0.25) || 1));
        
        return 0;
    }

#
# Child importing loop starts here
#
    do
    {

        $self->release_lock($implist_lock_handle) if defined $implist_lock_handle;

        my $max_grandchild = $self->l_max_grandchild();

        $implist_lock_handle = $self->get_lock(".lcgjm","jm-import.list",undef);
        $self->l_list_read("jm-import.list",\%import_list);

        my $start = time();

        my %changes;
        my %rl;

        my @keys_import_list = (keys %import_list);
        my $hash_len = scalar(@keys_import_list);
        $self->l_make_rl($hash_len,\%rl);

        my %TmpDescription;
        my %TmpENV;

        my $timeout = 0;
        my $nchild = 0;
        my %child_info;

        for(my $i=0;$i<$hash_len;$i++)
        {
            $self->release_lock($implist_lock_handle) if defined $implist_lock_handle;
            $implist_lock_handle = undef;

            my $next_import = $keys_import_list[$rl{$i}];

            my %TmpDescription;
            my %TmpENV;

            $child = fork();
            if (defined $child)
            {
                if ($child == 0)
                {
                    if ( -e $import_list{$next_import} )
                    {
                        eval `cat $import_list{$next_import}`;
                        %ENV = %TmpENV;

                        $self->{JobDescription} = new Globus::GRAM::JobDescription(\%TmpDescription);
                        $description = $self->{JobDescription};
                        $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};
                        $job_id = $description->jobid();

                        $status = $description->importstatus();
                        $batchout = $description->importbatchout();
                        $batcherr = $description->importbatcherr();
                        my $import_start_time = $description->importstarttime();

                        if ($next_import =~ /-acct$/)
                        {
                            my $curr_time = time();
                            $res = $self->l_save_acct_info(1);
                            exit(1) if ($res == -1 &&
                                       ($curr_time-$import_start_time<$MAX_ACCT_WAIT));
                        }
                        else
                        {
                            $self->l_import_cache($status,$batchout,$batcherr);
                        }
                    }
                    exit(0);
                }
                else
                {
                    $nchild++;
                    $child_info{$child}->{EID} = $next_import;
                }
            }

            my $last = 0;
            $timeout++ if (scalar(keys %changes)>0 && time()-$start>$MAX_IMPORTING_CYCLE_LEN);
            $last++ if $timeout || $i == ($hash_len-1);

            if ($nchild>=$max_grandchild || $last)
            {
                do
                {
                    my ($wpid,$status);
                    if ($nchild>0)
                    {
                        do
                        {
                            $wpid = waitpid(-1,0);
                        } while($wpid<=0);
                        $status = $?;
                        $nchild--;
                    }
                    if (defined $wpid)
                    {
                        my $entry_id = $child_info{$wpid}->{EID};
                        $changes{$entry_id} = 1 if (($status >> 8) == 0);
                        delete $child_info{$wpid};
                    }
                } while($nchild>=$max_grandchild || ($last && $nchild>0));
            }
            last if $last;
        }

        $self->l_my_sleep(10) if (scalar(keys %changes) == 0);

        $implist_lock_handle = $self->get_lock(".lcgjm","jm-import.list",undef) if !defined $implist_lock_handle;
        foreach my $entry_id (keys %changes)
        {
            $self->l_list_read_remove("jm-import.list",$entry_id,\%import_list);
        }
    } while(scalar(keys %import_list) > 0);

    $self->release_lock($external_lock_handle);
    $self->release_lock($implist_lock_handle);

    exit(0);
}

sub helper_armour
{
    my $self = shift;
    my ($text_ref) = @_;

    $$text_ref =~ s/\\/\\\\/g;
    $$text_ref =~ s/\$/\\\$/g;
    $$text_ref =~ s/`/\\\`/g;
}

sub helper_remove_tmp_file
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $file_in_tmp = "/tmp/submit.info.".$description->uniq_id();

    unlink($file_in_tmp);
}

sub helper_add_to_tmp_file
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $file_in_tmp = "/tmp/submit.info.".$description->uniq_id();
     
    my ($prefix,$array_ref) = @_;

    local(*FI);
    if (open(FI,">> ".$file_in_tmp))
    {
        foreach my $line (@$array_ref)
        {
            print FI $prefix.":".$line."\n";
        }
        close(FI);
    }
}

sub helper_get_from_tmp_file
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $file_in_tmp = "/tmp/submit.info.".$description->uniq_id();
     
    my ($prefix,$array_ref) = @_;

    local(*FI);
    if (open(FI,"< ".$file_in_tmp))
    {
        while(<FI>)
        {
            chomp(my $line=$_);
            if ($line =~ m|^$prefix:(.*)|)
            {
                push(@$array_ref,$1);
            }
        }
        close(FI);
    }
}

sub helper_remove_cache_export_dir
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag() || $ENV{'GLOBUS_GRAM_JOB_CONTACT'};
    my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";

    my $cache_url;
    my $cache_export_dir = $self->helper_get_cache_export_dir(\$cache_url);

    $self->cd_fork_and_exec_cmd(undef,0,0,$cache_pgm,"-delete","-t",$tag,$cache_url) if defined $cache_url;

    if (defined $cache_export_dir && $cache_export_dir ne "")
    {
        if (rename($cache_export_dir, $cache_export_dir.".rm"))
        {
            $self->cd_fork_and_exec_cmd(undef,0,0,"/bin/rm","-rf",$cache_export_dir.".rm");
        }
    }   
}

#
# Try to remove the cache export directory just before cleanuping
# up the gass cache on the gatekeeper. The cache export directory
# should be removed at the end of helper_cache_import(), but
# maybe import was never called.
#
sub cache_cleanup
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};

    $self->log("Entering Job Manager submit-helper implementation of cache_cleanup");

    my ($list_lock_handle) = $self->get_lock(".lcgjm","jm-cleanup.list",undef);
    my %cleanup_list;

    if (!$self->l_list_read_add("jm-cleanup.list",$tag,undef,undef,\%cleanup_list))
    {
        $self->release_lock($list_lock_handle);
        return {};
    }

    my %finished_list;
    my $fin_lock_handle = $self->get_lock(".lcgjm","jm-finished.list",undef);
    $self->l_list_read("jm-finished.list",\%finished_list);
    my %tmp_list = %finished_list;
    foreach my $tag_id (keys %tmp_list)
    {
        my $tag_base = $tag_id;
        $tag_base =~ s/-ts$//;

        if ($tag_id ne $tag_base)
        {
            $self->l_list_read_remove("jm-finished.list",$tag_id,\%finished_list) if !exists $tmp_list{$tag_base};
            next;
        }
        elsif (!exists $tmp_list{$tag_base."-ts"})
        {
            $self->l_list_read_add("jm-finished.list",$tag_id,undef,time(),\%finished_list);
        }

        next if exists $cleanup_list{$tag_id};

        my $fin_time = $finished_list{$tag_id."-ts"};
        my $fin_jd = $finished_list{$tag_id};
        my $uniqid;
    
        $fin_time =~ s/^PRIV://;

        my $now = time();

        if ($now>=$fin_time && $now-$fin_time>$MAX_FIN_TIME && -e $fin_jd)
        {
            my %TmpDescription;
            my %TmpENV;

            eval `cat $fin_jd`;
            $uniqid = $TmpDescription{uniqid}->[0];

            $self->l_tidy_gram_state($uniqid) if defined $uniqid;
            $self->l_list_read_add("jm-cleanup.list",$tag_id,$fin_jd,undef,\%cleanup_list);
            $self->l_list_read_remove("jm-finished.list",$tag_id,\%finished_list);
            $self->l_list_read_remove("jm-finished.list",$tag_id."-ts",\%finished_list);
        }
        else
        {
            if ( $now<$fin_time || !-e $fin_jd )
            {
                $self->l_list_read_remove("jm-finished.list",$tag_id,\%finished_list);
                $self->l_list_read_remove("jm-finished.list",$tag_id."-ts",\%finished_list);
            }
        }
    }
    $self->release_lock($fin_lock_handle);

    my $proc_lock_handle = $self->get_lock(".lcgjm","jm-cleanup.proc",0);

    if (!defined $proc_lock_handle)
    {
        $self->release_lock($list_lock_handle);
        return {};
    }

    $child = fork();
    if (!defined $child)
    {
        $self->release_lock($proc_lock_handle);
        $self->release_lock($list_lock_handle);
        return {};
    }

    if ($child == 0)
    {
        $self->l_daemonise();

        $self->chown_lock($proc_lock_handle);
        $self->chown_lock($list_lock_handle);
    }
    else
    {
        my ($pid1,$pid2);
        do
        {
            $pid1 = $self->getpid_lock($proc_lock_handle);
            $pid2 = $self->getpid_lock($list_lock_handle);
               
        } while(((defined $pid1 && $pid1 != $child) || (defined $pid2 && $pid2 != $child)) && ($self->l_my_sleep(0.25) || 1));
        
        return {};
    }

#
# Child cleanup loop starts here
#
    do
    {
        $self->release_lock($list_lock_handle);
        
        my @keys_cleanup_list = (keys %cleanup_list);
        my $hash_len = scalar(@keys_cleanup_list);
        my $next_clean = $keys_cleanup_list[int(rand($hash_len))];

        my %TmpDescription;
        my %TmpENV;

        if ( -e $cleanup_list{$next_clean} )
        {
            eval `cat $cleanup_list{$next_clean}`;
            %ENV = %TmpENV;

            $self->{JobDescription} = new Globus::GRAM::JobDescription(\%TmpDescription);
            $description = $self->{JobDescription};
            $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};

            my $uniq_id = $description->uniq_id();
            if ($uniq_id eq '')
            {
                my @Tmp = split( /\//, $tag );
                my $Id2 = pop @Tmp;
                my $Id1 = pop @Tmp;

                if ($Id1 && $Id2)
                {
                    $uniq_id = $Id1.".".$Id2;
                    $description->add('uniqid',$uniq_id);
                }
            }

            my $remote_io_file = ".lcgjm/.remote_io_ptr/remote_io_file-".$uniq_id;
            unlink($remote_io_file) if ($uniq_id ne '');
        
            $self->l_tidy_queues();
   
            $self->helper_remove_cache_export_dir();
            $self->helper_remove_tmp_file();
            $self->SUPER::cache_cleanup();

            my %finished_list;
            my $fin_lock_handle = $self->get_lock(".lcgjm","jm-finished.list",undef);
            $self->l_list_read_remove("jm-finished.list",$tag,\%finished_list);
            $self->l_list_read_remove("jm-finished.list",$tag."-ts",\%finished_list);
            $self->release_lock($fin_lock_handle);
        }

        $list_lock_handle = $self->get_lock(".lcgjm","jm-cleanup.list",undef);
        $self->l_list_read_remove("jm-cleanup.list",$next_clean,\%cleanup_list);

    } while(scalar(keys %cleanup_list) > 0);

    $self->release_lock($proc_lock_handle);
    $self->release_lock($list_lock_handle);

    exit(0);
}           

#
# Save the location of the directory in which the scratch directory should be
# made by the helper script.. The helper script will try
# to create a unique directory in this location on the batch
# worker.
#
sub make_scratchdir
{
    my $self = shift;
    my $description = $self->{JobDescription};

    my $scratch_prefix;
    my $scratch_suffix;
    my $dirname;

    $self->log("Entering Job Manager submit-helper implementation of make_scratchdir");
     
    $scratch_prefix = $description->scratch_dir_base();
    $scratch_suffix = $description->scratch_dir();

    if($scratch_suffix =~ m,^/,,)
    {
        $scratch_prefix = $scratch_suffix; 
    }
    elsif ($scratch_suffix !~ m,/$,,)
    {   
        $scratch_prefix .= "/$scratch_suffix";
    }
    else
    {
        $scratch_prefix .= $scratch_suffix;
    }
    
    $dirname = $scratch_prefix;

    my @scratch = ($dirname);
    $self->helper_add_to_tmp_file("scratch",\@scratch);
    
    $self->log("Using $dirname as the directory in which the scratch directory for this job will be made.");
    return {SCRATCH_DIR => $dirname}; 
}

#
# No scratchdir on the gatekeeper to remove
#
sub remove_scratchdir
{        
    my $self = shift;
    my $description = $self->{JobDescription};
    my $scratch_directory = $description->scratch_directory();

    $self->log("Executing Job Manager submit-helper implementation of remove_scratchdir (no action)");

    return {};
}

sub helper_init_cache_export_url
{
    my $self = shift;
    my ($cache_export_dir) = @_;
    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};
    my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";

    my $cache_export_url = "$tag/gk_cache_export_dir";
    $self->cd_fork_and_exec_cmd(undef,0,0,$cache_pgm,"-add","-t",$tag,"-n",$cache_export_url,"file:///dev/null");
    chomp(my $cache_export_ptr_file = `$cache_pgm -query -t $tag $cache_export_url`);

    if($cache_export_ptr_file ne "" && -e "$cache_export_ptr_file")
    {
        local(*CACHEPTR);
        open(CACHEPTR,"> ".$cache_export_ptr_file);
        print CACHEPTR $cache_export_dir."\n";
        close(CACHEPTR);
    }
}

sub helper_write_non_mpi_script
{
    my $self = shift;
    my ($script,$my_hostname,$local_x509,$x509_method,$cache_export_dir,$gram_scratch_dir,$emergx509,$exec_args) = @_;

    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};
    my $helper_pgm = $Globus::Core::Paths::libdir."/perl/Globus/GRAM/submit-helper.pl";

    $script->print("bootstrap=\`mktemp /tmp/bootstrap.XXXXXX\`; chmod 700 \$bootstrap\n");
    $script->print("cat >> \$bootstrap <<EOFbs\n");

    local(*HELPER);
    open(HELPER,"< ".$helper_pgm);
    while(<HELPER>)
    {
        chomp(my $line=$_);
        $self->helper_armour(\$line);
        $script->print($line."\n");
    }
    close(HELPER);
    if ($local_x509 ne '-')
    {
        if (lc($x509_method) eq 'x509self')
        {
            local(*X509);
            open(X509,"< $local_x509");
            while(<X509>)
            {
                chomp(my $line=$_);
                $self->helper_armour(\$line);
                $script->print("#X509#".$line."\n");
            }
            close(X509);
        }

        if (lc($x509_method) eq 'x509gpg' || lc($emergx509) eq 'pbsmsg')
        {
            my $passphrase = "";
            my @tmp = split('/',$cache_export_dir);
            my $fn = pop(@tmp);
            $fn .= ".gpg";

            for (my $i=0;$i<64;$i++)
            {
                $passphrase .= pack("C",97+rand(26));
            }

            if ($g_useGPG)
            {
                $script->print("#X509GPG#".$passphrase."\n");
                $self->write_gpg_file($local_x509,$cache_export_dir."/".$fn,$passphrase);
            }
            else
            {
                $self->cd_fork_and_exec_cmd(undef,0,0,"/bin/ln","-s",$local_x509,$cache_export_dir."/".$fn);
            }

            $x509_method = "X509GPG:".$fn if lc($x509_method) eq 'x509gpg';
        }
    }
    $script->print("EOFbs\n");
    
    if ($description->jobtype() eq 'multiple')
    {
        for(my $i = 0; $i < $description->count(); $i++)
        {
            $script->print("\$bootstrap ".$description->directory()." $my_hostname $local_x509 ".
                             $x509_method." ".$description->stdin()." ".$description->stdout()." stdoutftp ".$description->stderr()." stderrftp ".
                            "$cache_export_dir $tag ".$gram_scratch_dir." ".$emergx509." ".$description->executable()." $exec_args &\n");
        }
        $script->print("wait\n");
    }
    else
    {
        $script->print("\$bootstrap ".$description->directory()." $my_hostname $local_x509 ".
                         $x509_method." ".$description->stdin()." ".$description->stdout()." stdoutftp ".$description->stderr()." stderrftp ".
                        "$cache_export_dir $tag ".$gram_scratch_dir." ".$emergx509." ".$description->executable()." $exec_args\n");
    }
    $script->print("rm \$bootstrap\n");
}

sub helper_write_fake_mpi_script
{
    my $self = shift;
    my ($script,$rsh_env,$my_hostname,$local_x509,$x509_method,$cache_export_dir,$gram_scratch_dir,$emergx509,$exec_args) = @_;

    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};
    my $helper_pgm = $Globus::Core::Paths::libdir."/perl/Globus/GRAM/submit-helper.pl";

    $script->print("local_bootstrap=\`mktemp /tmp/local_bootstrap.XXXXXX\`; chmod 700 \$local_bootstrap\n");
    $script->print("cat >> \$local_bootstrap <<EOFrshbs\n");

    $script->print("#!/bin/sh\n");
    $script->print("$rsh_env\n");
    $script->print("bootstrap=\\\`mktemp /tmp/bootstrap.XXXXXX\\\`; chmod 700 \\\$bootstrap\n");
    $script->print("cat >> \\\$bootstrap <<EOFnombs\n");

    local(*HELPER);
    open(HELPER,"< ".$helper_pgm);
    while(<HELPER>)
    {
        chomp(my $line=$_);
        $self->helper_armour(\$line);
        $self->helper_armour(\$line);
        $script->print($line."\n");
    }
    close(HELPER);
    if ($local_x509 ne '-' && lc($x509_method) eq 'x509self')
    {
        local(*X509);
        open(X509,"< $local_x509");
        while(<X509>)
        {
            chomp(my $line=$_);
            $self->helper_armour(\$line);
            $self->helper_armour(\$line);
            $script->print("#X509#".$line."\n");
        }
        close(X509);
    }
    $script->print("EOFnombs\n");

    $script->print("\\\$bootstrap ".$description->directory()." $my_hostname $local_x509 ".
                       $x509_method." ".$description->stdin()." ".$description->stdout()." stdoutftp ".$description->stderr()." stderrftp ".
                      "$cache_export_dir $tag ".$gram_scratch_dir." ".$emergx509." ".$description->executable()." $exec_args\n");

    $script->print("rm \\\$bootstrap\n");
    $script->print("EOFrshbs\n");
}

sub write_gpg_file
{
    my $self = shift;
    my ($source,$dest,$passphrase) = @_;
    my ($res,$ret);

    my $cache_export_dir = $self->helper_get_cache_export_dir();
    return if !defined $cache_export_dir || $cache_export_dir eq "";

    pipe RH,WH;
    fcntl(RH, F_SETFD, 0);

    print WH $passphrase."\n";
    close(WH);

    $res = system("cat $source | gpg -c --batch --no-tty --no-options --homedir ".$cache_export_dir." ".
                  "--armor --passphrase-fd ".fileno(RH)." 2>/dev/null >$dest");
    $ret = $res >> 8;

    close(RH);
}

sub get_data_in_file
{
    my $self = shift;
    my ($file_dir,$file_name,$stamp_ref,$data_ref) = @_;

    my ($inode1,$inode2);
    local(*FILE);
    my $read;

    do
    {
        $read = 0;
        $inode1 = (stat($file_dir."/".$file_name))[1];

        if (open(FILE,"< ".$file_dir."/".$file_name))
        {
            @$data_ref = ();
            my $line_num = 0;
            while(<FILE>)
            {
                chomp(my $line = $_);
                $line_num++;
                if ($line_num == 1)
                {
                    $$stamp_ref = $line;
                }
                else
                {
                    push(@$data_ref,$line);
                }
            }
            close(FILE);
            $read++ if $line_num>0;
        }
        $inode2 = (stat($file_dir."/".$file_name))[1];
    } while(defined $inode1 && defined $inode2 && $inode1 != $inode2);

    return $read;
}

sub store_data_in_file
{
    my $self = shift;
    my ($file_dir,$file_name,$stamp,$data_ref) = @_;

    local(*FILE);
    my $written = 0;

    if (open(FILE,"> ".$file_dir."/".$file_name.".$$"))
    {
        print FILE $stamp."\n";
    
        foreach my $line (@$data_ref)
        {
            print FILE $line."\n";
        }
        close(FILE);
        rename $file_dir."/".$file_name.".$$", $file_dir."/".$file_name;
        $written++;
    }

    return $written;
}

sub proxy_update
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};

    $self->log("Entering Job Manager submit-helper implementation of proxy_update");
 
    my $cache_export_dir = $self->helper_get_cache_export_dir();
    return $self->SUPER::proxy_update(@_) if !defined $cache_export_dir || $cache_export_dir eq "";
    return $self->SUPER::proxy_update(@_) if (! -e $cache_export_dir."/.submit-script" || ! -e $cache_export_dir."/.emergency-x509");
  
    my @tmp = split('/',$cache_export_dir);
    my $gpg_file = pop(@tmp);
    $gpg_file .= ".gpg";

    my $passphrase;

    local(*SS);
    open(SS,"< ".$cache_export_dir."/.submit-script");
    while(<SS>)
    {
        chomp(my $line = $_);
        if ($line =~ /^#X509GPG#(\S+)$/)
        {
            $passphrase = $1;
        }
    }
    close(SS);

    if (defined $passphrase && $g_useGPG)
    {
        $self->write_gpg_file($cache_export_dir."/.emergency-x509",$cache_export_dir."/".$gpg_file,$passphrase);
        $self->log("Wrote gpg encoded copy of the renewed proxy");
    }

    return $self->SUPER::proxy_update(@_);
}

sub cmp_stamps
{
    my $self = shift;
    my ($a,$b) = @_;

    my ($a_time,$a_seq) = split(":",$a);
    my ($b_time,$b_seq) = split(":",$b);

    return 1 if ($a_time > $b_time);
    return -1 if ($a_time < $b_time);

    return 1 if ($a_seq > $b_seq);
    return -1 if ($a_seq < $b_seq);

    return 0;
}

sub stamp_to_time
{
    my $self = shift;
    my ($a) = @_;

    my ($a_time,$a_seq) = split(":",$a);

    return $a_time;
}

sub update_stamp
{
    my $self = shift;
    my ($old_stamp,$ctime) = @_;

    if (!defined $old_stamp || $self->stamp_to_time($old_stamp) != $ctime)
    {
        return $ctime.":0";
    }

    my ($a_time,$a_seq) = split(":",$old_stamp);
    $a_seq++;

    return $a_time.":".$a_seq;
}

sub make_a_poll_query
{
    my $self = shift;
    my ($file_dir,$file_name,$job_submit_time,$data_ref) = @_;
    my $read_stamp_fn = $file_dir."/poll-read.stamp";

    my $child;
    do
    {
        $child = waitpid(-1,WNOHANG);
    } until $child <= 0;

    local(*TS);
    if (open(TS,">> ".$read_stamp_fn))
    {
        close(TS);
        my $now = time();
        utime($now,$now,$read_stamp_fn);
    }

    my ($stamp,$exit_code,$ref_time,$age);

    {
        $exit_code = $self->get_data_in_file($file_dir,$file_name,\$stamp,$data_ref);
        $ref_time = time();

        if ($exit_code == 1)
        {
            $age = $ref_time - $self->stamp_to_time($stamp);
        }
        else
        {
            $stamp = undef;
            $age = "unknown";
            @$data_ref=();
        }

        my ($inconsistent,$check_proc) = (0,0);
        $inconsistent = 1 if ( defined $stamp && $self->stamp_to_time($stamp)>$ref_time );
        $check_proc = 1 if $inconsistent || !defined $stamp || $age > $MAX_POLL_AGE;

        if ($check_proc)
        {
            my $proc_lock_handle = $self->get_lock($file_dir,$file_name.".proc",0);

            if (defined $proc_lock_handle)
            {
                $self->log("Will start a batch system poll process in the background");
                $child = fork();

                if (!defined $child)
                {
                    $self->release_lock($proc_lock_handle);
                    return -1;
                }

                if ($child == 0)
                {
                    $self->l_daemonise();
                    $self->chown_lock($proc_lock_handle);

                    my $start_time = time();
                    my $ptime;
                    my $now;

                    do
                    {
                        $self->poll_batch_system($data_ref,\$ref_time);
    
                        my $new_stamp = $self->update_stamp($stamp,$ref_time);
                        $self->store_data_in_file($file_dir,$file_name,$new_stamp,$data_ref);

                        $self->release_lock($proc_lock_handle);
                        $proc_lock_handle = undef;

                        my $elapsed = time() - $ref_time;
                        $elapsed = 0 if $elapsed<0;
                        my $sleep_time = $MAX_POLL_AGE - $elapsed - 10;
                        $sleep_time = 0 if $sleep_time < 0;
                        $self->l_my_sleep($sleep_time);

                        $ptime = (CORE::stat($read_stamp_fn))[9];
                        $now = time();

                    } while( (!defined $ptime || ($ptime<=$now && $now-$ptime < $POLL_PROC_LIFE)) &&
                             ($start_time <= $now && $now-$start_time < $MAX_POLL_PROC_AGE) &&
                             defined($proc_lock_handle = $self->get_lock($file_dir,$file_name.".proc",0)) );

                    $self->release_lock($proc_lock_handle) if defined $proc_lock_handle;
                    exit(0);
                }
                else
                {
                    my $pid;
                    do
                    {
                        $pid = $self->getpid_lock($proc_lock_handle);
                    } while(((defined $pid && $pid != $child)) && ($self->l_my_sleep(0.25) || 1));
                }
            }

            if ($inconsistent)
            {
                $self->log("Current poll is inconsistent, waiting for next available poll");
                $self->l_my_sleep(10);
                redo;
            }
        }
    }

    if (!defined $job_submit_time || !defined $stamp ||
           $self->stamp_to_time($stamp)<=($job_submit_time+$LRMS_QUERY_DELAY))
    {
        $self->log("Cache too old for this job ($age): make_a_poll_query() returning 0");
        return 0;
    }
    
    $self->log("Using poll cache of age ".$age);
    return 1;
}

# The get_accounting_data() method should return a value indicating:
#
#   -2 Fatal error in gathering accounting information
#   -1 Temporary error in gathering accounting information
#    0 Accounting information successfuly saved
#
sub get_accounting_data
{
    my $self = shift;
    my ($data_ref) = @_;

    push(@$data_ref,"Sorry, no accounting information is collected from this type of batch system at the moment");

    return 0;
}

sub l_my_savedesc
{
    my $self = shift;
    my ($description,$filename) = @_;
    local(*OUT);                # protect

    if ( open( OUT, '>' . $filename.".".$$ ) )
    {
        print OUT '%TmpDescription = (', "\n";
        foreach my $var ( keys %{$description} )
        {
            my $value = $description->{$var};
            $value = [('/dev/null')] if $var eq 'logfile';

            print OUT '   \'', $var, '\' => ';
            $description->print_recursive( \*OUT, $value );
            print OUT ",\n";
        }
        print OUT ");\n";

        print OUT "%TmpENV = (\n";
        foreach my $var ( keys %ENV )
        {
            next if $var eq 'PWD';
            next if $var eq '_';

            print OUT '   \'', $var, '\' => ';
            print OUT '"'.quotemeta($ENV{$var}).'",'."\n";
        }
        print OUT ");\n";

        close(OUT);
        rename $filename.".".$$,  $filename;
    }
}

sub l_get_rand
{
    my $self = shift;

    my $chars;

    local(*RAND);
    open(RAND,"< /dev/urandom");
    sysread RAND,$chars,4;
    close(RAND);

    my $i=unpack('L',$chars);

    return $i;
}

sub l_import_cache
{
    my $self = shift;
    my ($status,$batchout,$batcherr) = @_;

    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag();
    my $job_id = $description->jobid();
    my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";

    $cache_export_dir = $self->helper_get_cache_export_dir();
    if (defined $cache_export_dir && $cache_export_dir ne "")
    {
        if ( -d $cache_export_dir)
        {
            my ($batchout_append,$batcherr_append) = (1,1);
            my ($batchout_name,$batcherr_name);

            $batchout_name = $batchout; 
            $batcherr_name = $batcherr; 

            if ($batchout_name =~ /^-(.*)/)
            {
                $batchout_name = $1;
                $batchout_append = 0;
            }
            elsif ($batchout_name =~ /^\./)
            {
                $batchout_name = $cache_export_dir.$batchout_name;
            }

            if ($batcherr_name =~ /^-(.*)/)
            {
                $batcherr_name = $1;
                $batcherr_append = 0;
            }
            elsif ($batcherr_name =~ /^\./)
            {
                $batcherr_name = $cache_export_dir.$batcherr_name;
            }

            $batchout_name = $cache_export_dir."/".$batchout_name if $batchout_name ne '' && $batchout_name !~ m|^/|;
            $batcherr_name = $cache_export_dir."/".$batcherr_name if $batcherr_name ne '' && $batcherr_name !~ m|^/|;
 
            local(*FL);
            open(FL,"< ".$cache_export_dir."/stdstreams.txt");
            chomp(my $ss_stdout = <FL>);
            chomp(my $ss_stderr = <FL>);
            close(FL);
     
            if ($batchout_name ne '' && -r "$batchout_name" )
            {
                l_append_file($batchout_name,$ss_stdout) if $batchout_append && $ss_stdout ne '';
                unlink($batchout_name);
            }
            if ($batcherr_name ne '' && -r "$batcherr_name" )
            {
                l_append_file($batcherr_name,$ss_stderr) if $batcherr_append && $ss_stderr ne '';
                unlink($batcherr_name);
            }

            local(*CACHEDIR);
            opendir(CACHEDIR,$cache_export_dir);
            foreach my $dir(readdir(CACHEDIR))
            {
                last if $status != Globus::GRAM::JobState::DONE;
                if ($dir =~ /import.txt.tar$/)
                {
                    $self->cd_fork_and_exec_cmd($cache_export_dir,0,0,"/bin/tar","xf",$dir) && next;
                    my $master_import_index = $dir;
                    $master_import_index =~ s/\.tar$//;

                    local(*IMPORT);
                    open(IMPORT,"< ".$cache_export_dir."/".$master_import_index) || next;
                    while(<IMPORT>)
                    {
                        chomp(my $line=$_);
                        my ($tag,$url,$file) = split('\s+',$line);

                        my %cache_url_tags;
#
# Avoid this query - does not scale well. It's ok that the
# tag will be created even if it already exists. In that case
# the gass utility silently ignores the create.
#
#                        $self->helper_get_url_and_tags("",$url,\%cache_url_tags);
#
#                        my @tags;
#                        @tags = split('\s+',$cache_url_tags{$url}) if defined($cache_url_tags{$url});
#    
#                        my $tag_exists = 0;
#                        foreach (@tags)
#                        {
#                            $tag_exists++ if $_ eq $tag;
#                        }

                        my $local;
                        chomp($local = `$cache_pgm -query -t $tag $url`);
                        if ($local eq "")
                        {
                            $self->cd_fork_and_exec_cmd(undef,0,0,$cache_pgm,"-add","-t",$tag,"-n",$url,"file:///dev/null");
                            chomp($local = `$cache_pgm -query -t $tag $url`);
                        }

                        next if $local eq "";
                        l_append_file($cache_export_dir."/".$file,$local);
                    }
                    close(IMPORT);
                }
            }
            closedir(CACHEDIR);
        }
        $self->helper_remove_cache_export_dir();
    }
}

sub l_daemonise
{
    my $self = shift;

    setsid();
    my $max_fd = sysconf(_SC_OPEN_MAX);

    my $ignore;
    if (exists $ENV{GATEKEEPER_ACCT_FD} &&
        exists $ENV{GATEKEEPER_JM_ID})
    {
        if ($ENV{GATEKEEPER_ACCT_FD} =~ /^(\d+)$/)
        {
            $ignore = $1;
        }
    }

    for(my $i=0;$i<$max_fd;$i++)
    {
        if (!defined $ignore || $i != $ignore)
        {
            local(*FH);
            open(FH,"<&=$i");
            close(FH);

            open(FH,">&=$i");
            close(FH);
        }
    }
    open(STDIN,"< /dev/null");
    open(STDOUT,"> /dev/null");
    open(STDERR,"> /dev/null");

    delete $self->{log};
}

sub l_list_read
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my ($list_name,$list_ref) = @_;

    my ($inode1,$inode2);
    local(*LIST);

    do
    {
        %$list_ref = ();
        $inode1 = (stat(".lcgjm/".$list_name))[1];
        if (open(LIST,"< .lcgjm/".$list_name))
        {
            while(<LIST>)
            {
                chomp(my $line = $_);
                if ($line =~ /^(\S+)\s+(\S+)$/)
                {
                    $list_ref->{$1} = $2;
                }
            }
            close(LIST);
        }
        $inode2 = (stat(".lcgjm/".$list_name))[1];
    } while(defined $inode1 && defined $inode2 && $inode1 != $inode2);

    return;
}

sub l_list_read_add
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my ($list_name,$entry_id,$copy_desc,$priv_data,$list_ref) = @_;
    my @unlink_list;

    return 0 if (defined $copy_desc && defined $priv_data);

    $self->l_list_read($list_name,$list_ref);

    if ( defined $priv_data )
    {
        push(@unlink_list,$list_ref->{$entry_id}) if exists $list_ref->{$entry_id} && $list_ref->{$entry_id} !~ /^PRIV:/;
        $list_ref->{$entry_id} = "PRIV:".$priv_data;
    }
    else
    {
        my $tmp;
        if ( !exists $list_ref->{$entry_id} || $list_ref->{$entry_id} =~ /^PRIV:/ )
        {
            chomp($tmp = `mktemp .lcgjm/jobdescription.XXXXXX`);
        }
        else
        {
            $tmp = $list_ref->{$entry_id};
        }

        if (defined $copy_desc)
        {
            $self->cd_fork_and_exec_cmd(undef,0,0,"/bin/cp",$copy_desc,$tmp);
        }
        else
        {
            $self->l_my_savedesc($description,$tmp);
        }
        $list_ref->{$entry_id} = $tmp;
    }

    mkdir ".lcgjm", 0700 if ! -d ".lcgjm";
    local(*LIST);
    if (open(LIST,"> .lcgjm/".$list_name.".$$"))
    {
        foreach my $key (keys %$list_ref)
        {
            print LIST $key." ".$list_ref->{$key}."\n";
        }
        close(LIST);
        rename ".lcgjm/".$list_name.".$$", ".lcgjm/".$list_name;

        unlink @unlink_list;

        return 1;
    }
    else
    {
        return 0;
    }
}

sub l_list_read_remove
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my ($list_name,$entry_id,$list_ref) = @_;
    my @unlink_list;

    $self->l_list_read($list_name,$list_ref);

    if (exists $list_ref->{$entry_id})
    {
        push(@unlink_list,$list_ref->{$entry_id}) if $list_ref->{$entry_id} !~ /^PRIV:/;
        delete $list_ref->{$entry_id};
    }

    if (scalar(keys %$list_ref)>0)
    {
        mkdir ".lcgjm", 0700 if ! -d ".lcgjm";
        local(*LIST);
        if (open(LIST,"> .lcgjm/".$list_name.".$$"))
        {
            foreach my $key (keys %$list_ref)
            {
                print LIST $key." ".$list_ref->{$key}."\n";
            }
            close(LIST);
            rename ".lcgjm/".$list_name.".$$", ".lcgjm/".$list_name;
        }
        else
        {
            return 0;
        }
    }
    else
    {
        unlink(".lcgjm/".$list_name);
    }

    unlink @unlink_list;
   
    return 1;
}

sub queue_submit
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $uniq_id = $description->uniq_id();
    my ($jm_type,$submit_arg) = @_;

    return undef if $self->l_check_memory();

    my $submit_id = time().":".$jm_type.":internal_".$self->l_get_rand().":".$uniq_id;

    $description->add('submitarg', $submit_arg);
    $description->add('jobid',$submit_id);

    mkdir '.lcgjm', 0700;
    mkdir '.lcgjm/.incoming-'.$jm_type, 0700;
    $self->l_my_savedesc($description,".lcgjm/.incoming-$jm_type/submit-".$uniq_id.".tmp");

    rename ".lcgjm/.incoming-$jm_type/submit-".$uniq_id.".tmp", ".lcgjm/.incoming-$jm_type/submit-".$uniq_id;

    return $submit_id;
}

sub get_loadavg
{
    my $self = shift;
    my ($loadavg_ref) = @_;

    my $stamp;
    my $now = time();
    my @data;
    my $read;
        
    $read = $self->get_data_in_file(".lcgjm","uptime.dat",\$stamp,\@data);
    if (!$read || $now-$stamp > $UPTIME_REFRESH || $now<$stamp)
    {
        my @load = (0.0, 0.0, 0.0);
        chomp(my $ut = `uptime`);
        @load = ($1, $2, $3) if $ut =~ /load average:\s+(\S+),\s+(\S+),\s+(\S+)/;
        @data=();
        push(@data,$load[0]." ".$load[1]." ".$load[2]);
        $self->store_data_in_file(".lcgjm","uptime.dat",$now,\@data);
    }

    @$loadavg_ref = split('\s+',$data[0]);
}

sub lookup_or_submit
{
    my $self = shift;
    my ($batch_id_ref,$submit_time_ref,$status_ref) = @_;
    my $description = $self->{JobDescription};
    my $entry_id = $description->jobid();

    my $now = time();

    if ($entry_id =~ /internal_/)
    {
        my ($queue_epoch,$jm_type,$iid,$iuqid) = split(":",$entry_id);
        my $running_gm = $self->l_is_gm();

        if ($running_gm != $$ && $now-$queue_epoch < 5)
        {
            $$status_ref = Globus::GRAM::JobState::PENDING;
            $$batch_id_ref = undef;
            $$submit_time_ref = undef;
            return;
        }
    }

    my $job_id;
    my $state;
    my $exit_code = $self->l_lookup_or_submit(\$job_id);

    if ($exit_code <= 0)
    {
        $job_id = undef;
        if ($exit_code == -2)
        {
            $state = Globus::GRAM::JobState::DONE;
        }
        elsif ($exit_code == -4)
        {
            $state = Globus::GRAM::JobState::FAILED;
        }
        else
        {
            $state = Globus::GRAM::JobState::PENDING;
        }
    }

    if (defined $state)
    {
        $$status_ref = $state;
    }
    else
    {
        $$status_ref = undef;
    }

    if (defined $job_id)
    {
        $$submit_time_ref = -1;
        $$batch_id_ref = $job_id;
        if ($$batch_id_ref =~ /^([^:]+):(.+)$/)
        {
            $$submit_time_ref = $1;
            $$batch_id_ref = $2;
        }
    }
    else
    {
        $$batch_id_ref = undef;
        $$submit_time_ref = undef;
    }

    return;
}

#
# l_lookup_or_submit() returns:
#
#      -4 == submission failed
#      -3 == fork error!!
#      -2 == job_id not found in submit queue
#      -1 == Job already in queue or child process already exists or could not get list lock to add entry
#       0 == started child process to do submitting
#       1 == job_id has been submitted, batch id put into jid_ref
#       2 == job_id does not seem to be an internal id. Copied it into jid_ref
#

sub l_lookup_or_submit
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my ($jid_ref) = @_;

    my $child;
    do
    {
        $child = waitpid(-1,WNOHANG); 
    } until $child <= 0;

    my $entry_id = $description->jobid();

    if ($entry_id !~ /internal_/)
    {
        $$jid_ref = $entry_id;
        return 2;
    }

    my $list_type = (split(":",$entry_id))[1];
    my $list_name = "jm-".$list_type."-submit.list";
    my $dq_name = "jm-".$list_type."-dequeue.list";
    my $proc_name = "jm-".$list_type."-submit.proc";
    my $lock_dir = ".lcgjm";

    my %dq_list;
    my %submit_list;
    my $list_lock_handle;

    {
        my $submit_entry = 0;
        $submit_entry++ if -e ".lcgjm/.incoming-$list_type/submit-".$description->uniq_id();

        $self->l_list_read($list_name,\%submit_list);
        $self->l_list_read($dq_name,\%dq_list);

        if (exists $submit_list{$entry_id})
        {
            if ($submit_list{$entry_id} =~ /^PRIV:(.*)/ && !exists $dq_list{$entry_id})
            {
                my $priv_data = $1;

                if ($priv_data =~ /^jid-(.*)/)
                {
                    $self->release_lock($list_lock_handle) if defined $list_lock_handle;

                    my $jid = $1;
                    if ($jid eq 'FAILED')
                    {
                        return -4;
                    }
                    else
                    {
                        $$jid_ref = $jid;
                        return 1;
                    }
                }
            }
        }
        else
        {
            if ( !$submit_entry && !exists $dq_list{$entry_id})
            {
                $self->release_lock($list_lock_handle) if defined $list_lock_handle;
                return -2;
            }

            if (!defined $list_lock_handle)
            {
                $list_lock_handle =  $self->get_lock($lock_dir,$list_name,0);
                $self->l_my_sleep(2.5 + 2.5*(rand(2) - 1.0)) if !defined $list_lock_handle;
                redo;
            }

            my %avail_submits;
            if (rand($INCOMING_DIR_CHECK) < 1)
            {
                local(*DIR);
                if (opendir(DIR,".lcgjm/.incoming-$list_type"))
                {
                    foreach my $entry (readdir(DIR))
                    {
                        if ($entry =~ /^submit-(\S+)/)
                        {
                            my $id = $1;
                            next if ($id =~ /\.tmp$/);

                            $avail_submits{$id} = ".lcgjm/.incoming-$list_type/".$entry;
                        }
                    }
                    closedir(DIR);
                }
            }
            else
            {
                $avail_submits{$description->uniq_id()} = ".lcgjm/.incoming-$list_type/submit-".$description->uniq_id();
            }

            my $dqueued = 0;
            foreach my $id (keys %avail_submits)
            {
                my %TmpDescription;
                my %TmpENV;
    
                eval `cat $avail_submits{$id}`;
                my $eid = $TmpDescription{jobid}->[0];

                if (exists $dq_list{$eid})
                {
                    my $save_description = $self->{JobDescription};
                    my %save_ENV = %ENV;

                    %ENV = %TmpENV;
                    $self->{JobDescription} = new Globus::GRAM::JobDescription(\%TmpDescription);
                    $description = $self->{JobDescription};

                    my $stage_inf = ".lcgjm/.incoming/stage_in-".$id;
                    unlink($stage_inf);
                    unlink($avail_submits{$id});
                    $self->l_list_read_remove($dq_name,$eid,\%dq_list);
                    $self->l_tidy_queues();

                    $dqueued++ if $eid eq $entry_id;
                    my $orig_pid = ($self->l_is_gm() != $$) ? getppid() : $$;
                    $self->write_to_gacct($orig_pid,"Job $eid has been REMOVED from internal submission queue (0)");

                    %ENV = %save_ENV;
                    $self->{JobDescription} = $save_description;
                    $description = $self->{JobDescription};
                }
                else
                {
                    $self->l_list_read_add($list_name,$eid."-jd",$avail_submits{$id},undef,\%submit_list);
                    $self->l_list_read_add($list_name,$eid,undef,"INCOMING-".$id,\%submit_list);
                    unlink $avail_submits{$id};
                }
            }    

            my %tmp_list = %dq_list;
            foreach my $internal_id (keys %tmp_list)
            {
                if (!exists $submit_list{$internal_id})
                {
                    $self->l_list_read_remove($dq_name,$internal_id,\%dq_list);
                    $dqueued++ if $internal_id eq $entry_id;
                }
            }

            if ( $dqueued )
            {
                $self->release_lock($list_lock_handle) if defined $list_lock_handle;
                return -2;
            }
        }
    }

    my $proc_lock_handle = $self->get_lock($lock_dir,$proc_name,0);
    if (!defined $proc_lock_handle)
    {
        $self->release_lock($list_lock_handle) if defined $list_lock_handle;
        return (exists $dq_list{$entry_id}) ? -2 : -1;
    }

    if (defined $list_lock_handle)
    {
        $self->release_lock($list_lock_handle);
        $list_lock_handle = undef;
    }

    my $orig_pid = ($self->l_is_gm() != $$) ? getppid() : $$;
    $child = fork();
    if (!defined $child)
    {
        $self->release_lock($proc_lock_handle);
        return -3;
    }

    if ($child == 0)
    {
        $self->l_daemonise();

        $self->chown_lock($proc_lock_handle);
    }
    else
    {
        my $pid1;
        do
        {
            $pid1 = $self->getpid_lock($proc_lock_handle);
               
        } while((defined $pid1 && $pid1 != $child) && ($self->l_my_sleep(0.25) || 1));

        return (exists $dq_list{$entry_id}) ? -2 : 0;
    }

#
# Child submit loop starts here
#
    my $none_left;

    do
    {
        $self->release_lock($list_lock_handle) if defined $list_lock_handle;

        my $max_grandchild = $self->l_max_grandchild();

        $list_lock_handle =  $self->get_lock($lock_dir,$list_name,undef);
        $self->l_list_read($list_name,\%submit_list);
        $self->l_list_read($dq_name,\%dq_list);

        my $start = time();
        $none_left = 1;

        my %changes;
        my %rl;

        my %tmp_list;
        my @keys_tmp_list;
        foreach my $key (keys %submit_list)
        {
            if (!exists $dq_list{$key})
            {
                $tmp_list{$key} = $submit_list{$key};
                push(@keys_tmp_list,$key);
            }
        }
        my $hash_len = scalar(@keys_tmp_list);
        $self->l_make_rl($hash_len,\%rl);

        my %TmpDescription;
        my %TmpENV;
        my @unlink_list;

        my $dq=0;

        my $timeout = 0;
        my $nchild = 0;
        my %child_info;

        for(my $ipass=0;$ipass<2 && $none_left;$ipass++)
        {
            for(my $i=0;$i<$hash_len;$i++)
            {
                my $entry_type = 0;
                my $entry_id = $keys_tmp_list[$rl{$i}];
                my $entry_jd;
                $entry_jd = $tmp_list{$entry_id."-jd"};

                if ($entry_id !~ /-jd$/ && defined $entry_jd && -e $entry_jd)
                {
                    if ($tmp_list{$entry_id} =~ /^PRIV:INCOMING-/)
                    {
                        $entry_type = 1;
                    }
                    elsif ($tmp_list{$entry_id} =~ /^PRIV:TOSUBMIT/ )
                    {
                        my ($queue_epoch,$jm_type,$iid,$iuqid) = split(":",$entry_id);
                        $entry_type = 2 if (time()-$queue_epoch > $MIN_SUBMIT_AGE);
                    }
                }

                if ($entry_type > 0)
                {
                    $self->release_lock($list_lock_handle) if defined $list_lock_handle;
                    $list_lock_handle = undef;

                    my ($rh,$wh);
                    {
                        local(*RH,*WH);
                        pipe RH,WH;
                        ($rh,$wh) = (*RH,*WH);
                    }

                    $child = fork();
                    if (defined $child)
                    {
                        if ($child == 0)
                        {
                            close($rh);

                            if ($entry_type == 2)
                            {
                                my $jid;

                                eval `cat $entry_jd`;
                                %ENV = %TmpENV;

                                $self->{JobDescription} = new Globus::GRAM::JobDescription(\%TmpDescription);
                                $description = $self->{JobDescription};

                                my $submitarg = $description->submitarg();
                                $jid = $self->submit_to_batch_system($submitarg);

                                print $wh "PRIV:jid-".$jid."\n";
                            }
                            elsif ($entry_type == 1)
                            {
                                my $uniq_id;
                                $uniq_id = $1 if ($tmp_list{$entry_id} =~ /^PRIV:INCOMING-(.*)/);

                                print $wh "PRIV:TOSUBMIT\n";
                            }
                            close($wh);
                            exit(0);
                        }
                        else
                        {
                            close($wh);
                            $nchild++;

                            my ($queue_epoch,$jm_type,$iid,$iuqid) = split(":",$entry_id);
                            $child_info{$child}->{EID} = $entry_id;
                            $child_info{$child}->{TYPE} = $entry_type;
                            $child_info{$child}->{RH} = $rh;
                            $child_info{$child}->{JMTYPE} = $jm_type;
                        }
                    }
                    else
                    {
                        close($rh);
                        close($wh);
                    }
                }

                $timeout++ if (scalar(keys %changes)>0 && time()-$start>$MAX_SUBMIT_CYCLE_LEN);

                $self->l_list_read($dq_name,\%dq_list);
                foreach my $key (keys %dq_list)
                {
                    $dq++ if exists $submit_list{$key};
                }

                $none_left = 0 if ($timeout || $dq);

                my $last = 0;
                $last++ if !$none_left || $i == ($hash_len-1);

                if ($nchild>=$max_grandchild || $last)
                {
                    do
                    {
                        my $wpid;
                        if ($nchild>0)
                        {
                            do
                            {
                                $wpid = waitpid(-1,0);
                            } while($wpid<=0);
                            $nchild--;
                        }
                        if (defined $wpid)
                        {
                            $entry_id = $child_info{$wpid}->{EID};

                            my $read_fh = $child_info{$wpid}->{RH};
                            chomp(my $reply=<$read_fh>);
                            close($read_fh);

                            $reply = "PRIV:jid-FAILED" if $reply eq '';

                            push(@unlink_list,$tmp_list{$entry_id}) if ($reply ne $tmp_list{$entry_id} && $tmp_list{$entry_id} !~ /^PRIV:/);
                            $tmp_list{$entry_id} = $reply;

                            $changes{$entry_id}->{DATA} = $reply;
                            $changes{$entry_id}->{TYPE} = $child_info{$wpid}->{TYPE};
                            $changes{$entry_id}->{JMTYPE} = $child_info{$wpid}->{JMTYPE};

                            delete $child_info{$wpid};
                        }
                    } while($nchild>=$max_grandchild || ($last && $nchild>0));
                }
                last if $last;
            }
        }

        $list_lock_handle = $self->get_lock($lock_dir,$list_name,undef) if !defined $list_lock_handle;
        $self->l_list_read($list_name,\%submit_list);

        if (scalar(keys %changes)>0)
        {
            $none_left = 0;
            foreach my $entry_id (keys %changes)
            {
                my $entry_type = $changes{$entry_id}->{TYPE};
                my $data = $changes{$entry_id}->{DATA};
                my $jm_type = $changes{$entry_id}->{JMTYPE};

                if ($entry_type == 1)
                {
                    if ($data =~ /^PRIV:jid-FAILED/)
                    {
                        $self->l_list_read_add($list_name,$entry_id,undef,"jid-FAILED",\%submit_list);
                        $self->write_to_gacct($orig_pid,"Job $entry_id FAILED during stage in");
                    }
                    else
                    {
                        $self->l_list_read_add($list_name,$entry_id,undef,"TOSUBMIT",\%submit_list);
                    }
                }
                elsif ($entry_type == 2)
                {
                    if ($data =~ /^PRIV:jid-(.*)/)
                    {
                        my $jid = $1;
                        if ($jid eq "FAILED")
                        {
                            $self->l_list_read_add($list_name,$entry_id,undef,"jid-FAILED",\%submit_list); 
                            $self->write_to_gacct($orig_pid,"Job $entry_id FAILED during submission to batch system $jm_type");
                        }
                        else
                        {
                            $self->l_list_read_add($list_name,$entry_id,undef,"jid-".time().":".$jid,\%submit_list); 
                            $self->write_to_gacct($orig_pid,"Submitted job $entry_id to batch system $jm_type with ID $jid");
                        }
                    }
                }
            }
        }
        unlink @unlink_list;

        my %import_list;
        $self->l_list_read($dq_name,\%dq_list);
        my $retry;

        do
        {
            $retry = 0;
            %tmp_list = %dq_list;
            my $count = 0;

            foreach my $entry_id (keys %tmp_list)
            {
                $count++;
                $self->l_list_read("jm-import.list",\%import_list);
                if (exists $import_list{$entry_id})
                {
                    # NB We have ignored the 'acct' entries in the import list
                    $retry++;
                    next;
                }

                my $submit_jd;
                $submit_jd = $submit_list{$entry_id."-jd"};
    
                if (exists $submit_list{$entry_id} &&
                    defined $submit_jd && -e $submit_jd)
                {
                    my $batch_id;
                    my $prev_state = $submit_list{$entry_id};

                    eval `cat $submit_jd`;
                    %ENV = %TmpENV;

                    $self->{JobDescription} = new Globus::GRAM::JobDescription(\%TmpDescription);
                    $description = $self->{JobDescription};

                    if ($prev_state =~ /^PRIV:jid-(.*)/)
                    {
                        my $job_id = $1;
                        $batch_id = $job_id;
                        if ($batch_id =~ /^([^:]+):(.+)$/)
                        {
                            $batch_id = $2;
                        }
                    }
                    elsif ($prev_state =~ /^PRIV:INCOMING-(.*)/)
                    {
                        my $uniq_id = $1;
                        my $stage_inf = ".lcgjm/.incoming/stage_in-".$uniq_id;
                        unlink $stage_inf;
                    }

                    if (defined $batch_id && $batch_id ne "FAILED")
                    {
                        my $cancel_ret = $self->cancel_in_batch_system($batch_id);
                        if ($cancel_ret == Globus::GRAM::Error::JOB_CANCEL_FAILED)
                        {
                            $self->write_to_gacct($orig_pid,"Error: Job $entry_id (batch ID $batch_id) NOT REMOVED from batch system");
                        }
                        else
                        {
                            $self->write_to_gacct($orig_pid,"Job $entry_id (batch ID $batch_id) REMOVED from batch system ok");
                        }
                    }
                    else
                    {
                        $self->write_to_gacct($orig_pid,"Job $entry_id has been REMOVED from internal submission queue ($prev_state)");
                    }

                    $self->l_list_read_remove($dq_name,$entry_id,\%dq_list);
                    $self->l_tidy_queues();
                }
            }
            sleep 5 if $retry;
        } while($retry);

        %tmp_list = %submit_list;
        foreach my $entry_id (keys %tmp_list)
        {
            if ($entry_id =~ /-jd$/)
            {
                my $eid = $entry_id;
                $eid =~ s/-jd$//;

                if (!exists $tmp_list{$eid})
                {
                    $self->l_list_read_remove($list_name,$entry_id,\%submit_list);
                }
            }
            else
            {
                my $submit_jd;
                $submit_jd = $submit_list{$entry_id."-jd"};

                if (!(defined $submit_jd && -e $submit_jd))
                {
                    $self->l_list_read_remove($list_name,$entry_id,\%submit_list);
                }
            }
        }

    } while( !$none_left );

    $self->release_lock($proc_lock_handle);
    $self->release_lock($list_lock_handle);

    exit(0);
}            

sub l_tidy_queues
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $uniq_id = $description->uniq_id();
    my $internal_id = $description->jobid();
    my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};

    my %running;
    my $running_lock_handle = $self->get_lock(".lcgjm","jm-running.list",undef);
    $self->l_list_read_remove("jm-running.list",$uniq_id,\%running);
    $self->release_lock($running_lock_handle);

    my %finished_list;
    my $fin_lock_handle = $self->get_lock(".lcgjm","jm-finished.list",undef);
    $self->l_list_read_add("jm-finished.list",$tag,undef,undef,\%finished_list);
    $self->l_list_read_add("jm-finished.list",$tag."-ts",undef,time(),\%finished_list);
    $self->release_lock($fin_lock_handle);

    if ($internal_id =~ /internal_/)
    {
        my ($queue_epoch,$jm_type,$iid,$iuqid) = split(":",$internal_id);

        my %submit_list;
        my %dq_list;
        my $list_lock_handle = $self->get_lock(".lcgjm","jm-".$jm_type."-submit.list",undef);
        $self->l_list_read_remove("jm-".$jm_type."-submit.list",$internal_id,\%submit_list);
        $self->l_list_read_remove("jm-".$jm_type."-submit.list",$internal_id."-jd",\%submit_list);
        $self->release_lock($list_lock_handle);
    }
}

sub write_to_gacct
{
    my $self = shift;
    my ($pid,$info) = @_;

    my $logged=0;

    if (exists $ENV{GATEKEEPER_ACCT_FD} &&
        exists $ENV{GATEKEEPER_JM_ID})
    {
        my $acct_fd;
        if ($ENV{GATEKEEPER_ACCT_FD} =~ /^(\d+)$/)
        {
            $acct_fd = $1;
        }

        if (defined $acct_fd)
        {
            if (open(ACCTFD,">&=".$acct_fd))
            {
                print ACCTFD $ENV{GATEKEEPER_JM_ID}." ".$info."\n";
                $logged++;
            }
        }
    }

#    if (!$logged)
    {
#        my $jm_id = "NONE";
#        $jm_id = $ENV{GATEKEEPER_JM_ID} if exists $ENV{GATEKEEPER_JM_ID};

        $self->cd_fork_and_exec_cmd(undef,0,0,"/usr/bin/logger","-t","gridinfo","--","[$pid"."-"."$$] ".$info);
    }
}

sub rewrite_urls
{
    my $self = shift;
    my $description = $self->{JobDescription};
 
    $self->log("Entering Job Manager submit-helper implementation of rewrite_urls");

#    my ($lock_handle) = $self->get_lock(".lcgjm","rewrite_urls",undef);
    my $result = $self->SUPER::rewrite_urls(@_);
#    $self->release_lock($lock_handle);

    $self->log("Leaving Job Manager submit-helper implementation of rewrite_urls");

    return $result;
}

sub stage_out
{
    my $self = shift;
    my $description = $self->{JobDescription};
 
    $self->log("Entering Job Manager submit-helper implementation of stage_out");

#    my ($lock_handle) = $self->get_lock(".lcgjm","stage_out",undef);
    my $result = $self->SUPER::stage_out(@_);
#    $self->release_lock($lock_handle);

    $self->log("Leaving Job Manager submit-helper implementation of stage_out");

    return $result;
}

sub proxy_relocate
{
    my $self = shift;
    my $description = $self->{JobDescription};

    my $result = $self->SUPER::proxy_relocate(@_);

    return $result;
}

sub remote_io_file_create
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $uniq_id = $description->uniq_id();

    $self->log("Entering Job Manager submit-helper implementation of remote_io_file_create");

    mkdir '.lcgjm', 0700;
    mkdir '.lcgjm/.remote_io_ptr', 0700;

    my $path = $ENV{HOME}."/.lcgjm/.remote_io_ptr/remote_io_file-".$uniq_id;

    local(*FH);
    if (open(FH,"> ".$path))
    {
        print FH $description->remote_io_url, "\n";
        close(FH);
    }

    $self->log("Leaving Job Manager submit-helper implementation of remote_io_file_create");
    return { REMOTE_IO_FILE => $path };
}

sub stage_in
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $uniq_id = $description->uniq_id();

    mkdir '.lcgjm', 0700;
    mkdir '.lcgjm/.incoming', 0700;

    # Must stage_in during the globus job request;
    # stage in files may be unavailable or modified later
    $self->l_stage_in();
    return {};
}

#
# The file_stage_in list is only present in the job description until stage_in,
# after successful stage_in the list elements are removed. Therefore build a list at
# the time stage_in() is called.
#
sub l_stage_in
{
    my $self = shift;
    my $description = $self->{JobDescription};
 
    $self->log("Entering Job Manager submit-helper implementation of stage_in");

    my @stage_in;
    my @stage_in_shared;
    foreach ($description->file_stage_in())
    {
        next unless defined $_;
        my ($remote, $local) = ($_->[0], $_->[1]);
        push(@stage_in,$remote." ".$local);
    }
    foreach ($description->file_stage_in_shared())
    {
        next unless defined $_;
        my ($remote, $local) = ($_->[0], $_->[1]);
        push(@stage_in_shared,$remote." ".$local);
    }

    $self->helper_add_to_tmp_file("stage_in",\@stage_in);
    $self->helper_add_to_tmp_file("stage_in_shared",\@stage_in_shared);

    my $result = $self->SUPER::stage_in(@_);

    $self->log("Leaving Job Manager submit-helper implementation of stage_in");

    return 0;
}

sub l_make_rl
{
    my $self = shift;
    my ($len,$hash_ref) = @_;

    %$hash_ref = ();

    for(my $i=0;$i<$len;$i++)
    {
        my $j;
        do
        {
            $j = int(rand($len));
        } while(exists $hash_ref->{$j});
        $hash_ref->{$j} = $i;
    }   
}

sub cancel
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $internal_id = $description->jobid();
    my $uniq_id = $description->uniq_id();
    my $orig_pid = ($self->l_is_gm() != $$) ? getppid() : $$;

    if ($internal_id =~ /internal_/)
    {
        my ($queue_epoch,$jm_type,$iid,$iuqid) = split(":",$internal_id);

        my $list_lock_handle =  $self->get_lock(".lcgjm","jm-".$jm_type."-submit.list",undef);

        my %dq_list;
        $self->l_list_read_add("jm-".$jm_type."-dequeue.list",$internal_id,undef,"DEQUEUE",\%dq_list);
        $self->release_lock($list_lock_handle);
        $self->write_to_gacct($orig_pid,"Job $internal_id added to DEQUEUE list");

        my ($batch_id,$submit_time,$status);
        my $attempts=0;
        do
        {
            $self->lookup_or_submit(\$batch_id,\$submit_time,\$status);

            $self->l_list_read("jm-".$jm_type."-dequeue.list",\%dq_list);
            sleep 10 if exists $dq_list{$internal_id}; #### MODIFIED ####

            $attempts++;
        } while(exists $dq_list{$internal_id} && $attempts<12);
        
        return Globus::GRAM::Error::JOB_CANCEL_FAILED if exists $dq_list{$internal_id};
       
        return { JOB_STATE => Globus::GRAM::JobState::FAILED };
    }
    else
    {
        my $batch_id = $internal_id;

        my $cancel_ret = $self->cancel_in_batch_system($batch_id);

        if ($cancel_ret == Globus::GRAM::Error::JOB_CANCEL_FAILED)
        {
            $self->write_to_gacct($orig_pid,"Job $internal_id (batch ID $batch_id) NOT REMOVED from batch system");
        }
        else
        {
            $self->write_to_gacct($orig_pid,"Job $internal_id (batch ID $batch_id) REMOVED from batch system ok");
        }

        return $cancel_ret;
    }
}

sub l_check_memory
{
    my $self = shift;

    local(*MEMINFO);
    my ($memtot,$memfree,$swaptot,$swapfree);

    if (open(MEMINFO,"< /proc/meminfo"))
    {
        while(<MEMINFO>)
        {
            chomp(my $line = $_);
            $memtot = $1 if ($line =~ /^MemTotal:\s*(\d+)\s*kB/);
            $memfree = $1 if ($line =~ /^MemFree:\s*(\d+)\s*kB/);
            $swaptot = $1 if ($line =~ /^SwapTotal:\s*(\d+)\s*kB/);
            $swapfree = $1 if ($line =~ /^SwapFree:\s*(\d+)\s*kB/);
        }
        close(MEMINFO);
    }
    
    return 0 if !(defined $memtot && defined $memfree && defined $swaptot && defined $swapfree);
    return 0 if ($memtot+$swaptot <= 0 || $memfree+$swapfree<0 );

    my $freefrac = ($memfree+$swapfree)/($memtot+$swaptot);

    return 1 if $freefrac < $MIN_MEM_FREE;
    return 0;
}

sub l_my_sleep
{
    my $self = shift;
    my ($wait_time) = @_;

    my $nfound;
    my $start = time();
    my $w = $wait_time;
    my $error;

    do
    {
        $nfound = select( undef,undef,undef,$w);
        $error = $!;
        $w = $wait_time - (time()-$start);
    } while($nfound<0 && $error =~ /Interrupted/ && $w>0);
}

sub l_delay_parent
{
    my $self = shift;
    my ($ppid,$cur_load) = @_;
    my $description = $self->{JobDescription};

    my $init_prob = ($cur_load - $PARENT_PAUSE_MIN_LOAD)/$PARENT_PAUSE_LOAD_RANGE;
    $init_prob = (($init_prob+0.5)*($init_prob+0.5)-0.25)/2 if $init_prob>0;
    return if rand(1) >= $init_prob;

    my $list_lock_handle = $self->get_lock(".lcgjm","jm-suspend.list",undef);
    my %suspend_list;

    kill 19, $ppid;

    if (!$self->l_list_read_add("jm-suspend.list",$ppid,undef,time(),\%suspend_list))
    {
        $self->release_lock($list_lock_handle);
        kill 18, $ppid;
        return;
    }

    my $proc_lock_handle = $self->get_lock(".lcgjm","jm-suspend.proc",0);

    if (!defined $proc_lock_handle)
    {
        $self->release_lock($list_lock_handle);
        return;
    }

    $child = fork();
    if (!defined $child)
    {
        $self->release_lock($proc_lock_handle);
        $self->release_lock($list_lock_handle);
        return;
    }

    if ($child == 0)
    {
        $self->l_daemonise();

        $self->chown_lock($proc_lock_handle);
        $self->chown_lock($list_lock_handle);
    }
    else
    {
        my ($pid1,$pid2);
        do
        {
            $pid1 = $self->getpid_lock($proc_lock_handle);
            $pid2 = $self->getpid_lock($list_lock_handle);
               
        } while(((defined $pid1 && $pid1 != $child) || (defined $pid2 && $pid2 != $child)) && ($self->l_my_sleep(0.25) || 1));
        
        return;
    }

#
# Child suspend watching starts here
#
    do
    {
        $self->release_lock($list_lock_handle) if defined $list_lock_handle;
        $list_lock_handle = undef;

        my %hash;
        foreach my $key (keys %suspend_list)
        {
            $hash{$key} = $suspend_list{$key};
            $hash{$key} =~ s/^PRIV://;
        }
        my @sorted_list = sort { $hash{$a} <=> $hash{$b} } keys %hash;
        my $next_ppid = $sorted_list[0];
        my $suspend_time = $hash{$next_ppid};
        my $now = time();

        my @loadavg;
        $self->get_loadavg(\@loadavg);
        my $prob = ($loadavg[0] - $PARENT_PAUSE_MIN_LOAD)/$PARENT_PAUSE_LOAD_RANGE;
        $prob = (($prob+0.5)*($prob+0.5) - 0.25)/2 if $prob>0;

        if (rand(1) < $prob && $now-$suspend_time < $MAX_PAUSE_PERIOD && $suspend_time<=$now)
        {
            $self->l_my_sleep(1);
        }
        else
        {
            kill 18, $next_ppid;

            $self->l_my_sleep(1);
            $list_lock_handle = $self->get_lock(".lcgjm","jm-suspend.list",undef);
            $self->l_list_read_remove("jm-suspend.list",$next_ppid,\%suspend_list);
        }

    } while(scalar(keys %suspend_list) > 0);

    $self->release_lock($proc_lock_handle);
    $self->release_lock($list_lock_handle);

    exit(0);
}

sub l_max_grandchild
{
    my $self = shift;

    my $max_grandchild;
    my @loadavg;
    do {
        $self->get_loadavg(\@loadavg);
        $max_grandchild = ($MAX_GRANDCHILD_LOAD - $loadavg[0])/$MAX_GRANDCHILD_SFACT;
        $max_grandchild = int($max_grandchild);
        $max_grandchild = 3 if $max_grandchild>3;
        $max_grandchild = 1 if $max_grandchild<1;
        sleep $UPTIME_REFRESH if $max_grandchild < 1;
    } while($max_grandchild<1);

    return $max_grandchild;
}

sub l_is_gm
{
    my $self = shift;

    my $globus_location = "/opt/globus";
    $globus_location = $ENV{GLOBUS_LOCATION} if exists $ENV{GLOBUS_LOCATION};
    my $monitor_lock = $globus_location."/tmp/grid_manager_monitor_agent_log.$<.lock";
    my ($LockPid,$LockTime);
    local(*LOCK);

    my $retry_open;
    do
    {
        $retry_open = 0;
        if ( open( LOCK, $monitor_lock ) )
        {
            flock(LOCK,LOCK_SH);
            my $ino = (CORE::stat($monitor_lock))[1];
            if (!defined $ino || $ino != (fstat(fileno(LOCK)))[1])
            {
                flock(LOCK,LOCK_UN);
                close(LOCK);
                $retry_open++;
            }
        }
        else
        {
            return 0;
        }
    } while($retry_open);

    my $running_gm = 0;
    my $LockMtime = (CORE::stat( $monitor_lock ))[9];

    while(<LOCK>)
    {
        chomp(my $line= $_ );
        if ( $line =~ /(\d+) (\d+)/ )
        {
            $LockPid = $1;
            $LockTime = $2;
        }
    }

    flock(LOCK,LOCK_UN);
    close( LOCK );

    $running_gm = $LockPid if (defined $LockPid && defined $LockTime && time()-$LockMtime<120);

    return $running_gm;
}

#
# NB When l_tidy_gram_state is called the job description does not
#    necessarily correspond to job specified by id
#
sub l_tidy_gram_state
{
    my $self = shift;
    my ($id) = @_;

    my $globus_location = "/opt/globus";
    $globus_location = $ENV{GLOBUS_LOCATION} if exists $ENV{GLOBUS_LOCATION};

    chomp(my $hostname = `hostname -f`);

    my $sf = $globus_location."/tmp/gram_job_state/job.".$hostname.".".$id;

    unlink($sf);
    unlink($sf.".lock");
}

#
# NB When l_exists_gram_state is called the job description does not
#    necessarily correspond to job specified by id
#
sub l_exists_gram_state
{
    my $self = shift;
    my ($id) = @_;

    my $globus_location = "/opt/globus";
    $globus_location = $ENV{GLOBUS_LOCATION} if exists $ENV{GLOBUS_LOCATION};

    chomp(my $hostname = `hostname -f`);

    my $sf = $globus_location."/tmp/gram_job_state/job.".$hostname.".".$id;

    return 1 if -e $sf;

    return 0;
}

sub cd_fork_and_exec_cmd
{
    my $self = shift;
    my $dir = shift;
    my $stdout = shift;
    my $stderr = shift;

    my $pid = fork();
    return undef unless defined $pid;

    if ( $pid == 0 )
    {
       # child
       $SIG{INT} = 'IGNORE';
       $SIG{QUIT} = 'IGNORE';

       if ($stdout == 0)
       {
           open STDOUT, '>>/dev/null';
       } elsif ($stdout == 2)
       {
           open STDOUT, '>&STDERR';
       }

       if ($stderr == 0)
       {
           open STDERR, '>>/dev/null';
       } elsif ($stderr == 1)
       {
           open STDERR, '>&STDOUT';
       }

       chdir $dir if defined $dir;
       exec { $_[0] } @_;
       exit 127;
    }   

    # parent
    my ($error,$ret,$status);
    do
    {
        $ret = waitpid($pid,0);
        $status = $?;
        $error = $!;
    } while($ret==-1 && $error =~ /Interrupted/);

    return $status;
}

sub l_append_file
{
    my ($src,$dest) = @_;
    my $buff;

    local(*SRC);
    local(*DEST);

    if (open(SRC,"< ".$src))
    {
        if (open(DEST,">> ".$dest))
        {
            binmode(SRC);
            binmode(DEST);
 
            while (read(SRC, $buff, 128 * 2**10))
            {
                print DEST $buff;
            }
            close(DEST);
        }
        close(SRC);
    }
}

# Returns:
#
#   -3 Fatal error in saving accounting information
#   -2 Fatal error in gathering accounting information
#   -1 Temporary error in gathering accounting information
#    0 Accounting information successfuly saved
#    1 Job already appears to have accounting information saved
#      and 'rewrite' not set
#
sub l_save_acct_info
{
    my $self = shift;
    my ($rewrite) = @_;

    my $description = $self->{JobDescription};
    my $tag = $description->cache_tag();
    my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";

    my $acct_url = "$tag/job-acct.info";
    my $acct_file_name = `$cache_pgm -query -t $tag $acct_url`;   
    chomp($acct_file_name);

    if ($acct_file_name ne '')
    {
        return 1 if !$rewrite;
    }
    else
    {
        $acct_file_name = "/tmp/jm-acct-tmp.$$";
        $rewrite = 0;
    }

    local(*CF);
    if (open(CF,"> ".$acct_file_name))
    {
        my @data;
        my $res = $self->get_accounting_data(\@data);

        foreach my $line (@data)
        {
            print CF $line."\n";
        }
        close(CF);
        if (!$rewrite)
        {
            $self->cd_fork_and_exec_cmd(undef,0,0,$cache_pgm,"-add","-t",$tag,"-n",$acct_url,"file:".$acct_file_name);
            unlink($acct_file_name);
        }
        return $res;
    }

    return -3;
}


1;
