#!/usr/bin/perl -w
use strict;
use Fcntl;

#    open(OLDERR11, ">&STDERR");close(OLDERR11);
#    open(STDERR, "> /tmp/frederr.1");
#    select(STDERR); $| = 1;     # make unbuffered
#    open(OLDOUT11, ">&STDOUT");close(OLDOUT11);
#    open(STDOUT, "> /tmp/fredout.1");
#    select(STDOUT); $| = 1;     # make unbuffered

#
# submit-helper.pl is a wrapper around the globus job executable
# that aims to remove any assumptions of a shared file system between the
# gatekeeper and the batch worker that the job executes on. We require
# GSIFTP to transfer files to or from the gatekeeper.
#
# Below is a description of the helper script's required arguments. Arguments
# that are interpreted as a relative path on the gatekeeper are prepended with
# the GRAMI_DIRECTORY, whereas relative paths on the batch node are prepended
# with the inital working directory for the job.
#
# Arguments: GRAMI_DIRECTORY
#                             The directory specified by the 'directory' directive
#                             in the job RSL. If this is something other than the home
#                             directory of the user we are running as and that directory
#                             exists on the batch worker, that will be used for the job's
#                             initial working directory. If GRAMI_DIRECTORY doesn't exist
#                             on the batch worker or it is the home directory of the user,
#                             the initial working directory for the job will be the initial
#                             directory in which the helper is started.
#
#            GATE_HOSTNAME
#                             The hostname of the gatekeeper machine.
#
#            REMOTE_X509_NAME Can be either '-', if X509 is not being used, or is the path
#                             name of this job's X509 proxy certificate on the gatekeeper
#                             machine. The current proxy is used until it is almost expired,
#                             at which point a fresh proxy is assumed to have been placed
#                             into this file on the gatekeeper, so it is fetched and the current
#                             proxy replaced.
#
#            INITAL_X509      (Only honored if REMOTE_X509_NAME is not '-')
#
#                             One of:
#
#                            X509SELF     Means that the inital X509 proxy
#                                         certificate for this job is to extracted
#                                         from the helper script. Lines with the prefix
#                                         '#X509#' are taken and the prefix removed.
#
#                            X509BATCH:fn Means that the inital X509 proxy should
#                                         have been placed in the inital working
#                                         directory of the helper script and have the
#                                         filename 'fn'. If the file is not found in the
#                                         initial working directory an attempt is made to
#                                         open it in the home directory of the user on
#                                         this node. If 'fn' is '-' then STDIN is used.
#                                         If a file is specified it is removed after it is read.
#
#                            X509GPG:fn   The initial X509 proxy is encrypted in the file
#                                         'fn', relative to the initial working directory of
#                                         the job. If the specified file is not found an attempt
#                                         is made to open it relative to the home directory of
#                                         the user on this node. The file fn is deleted after
#                                         it is read. The passphase is found appened to the
#                                         submission script.
#
#                            NONE         No initial X509 proxy it to be used. This means
#                                         that there should be an 'emergency' method
#                                         to get an X509 proxy. See EMERGX509 below.
#
#            JOB_STDIN        Can be either '-' or a pathname/URL of a file.
#
#                             If '-' is used then the job executable is started
#                             with the current STDIN (ie the STDIN proved by the batch system).
#
#                             If an absolute pathname is supplied and if it that file exists
#                             on the batch worker that file will be used as STDIN for the job. 
#                             If JOB_STDIN is a URL that URL is lookedup in the local
#                             gass cache, if the URL exists the corresponding file in the cache
#                             is used. If CACHE_EXPORT_DIR (see later) is being used and
#                             JOB_STDIN was the path to a cache entry on the gatekeeper that path
#                             is translated to the gass cache entry copied over to the local cache.
#                             
#                             Otherwise if the file is not available localy on the batch worker, an
#                             attempt is made to retieve the file from the gatekeeper.
#
#            REMOTE_STDOUT    Can be either '-' or the pathname of the file to which stdout
#                             from the job will be recorded on the gatekeeper. In case the
#                             location is specified, the STDOUT_XFER argument determines
#                             how the stdout is to be transfered back to the gatekeeper.
#
#            STDOUT_XFER      (Only honored if REMOTE_STDOUT is not '-')
#
#                             If stdout_xfer is:
#
#                               STDOUTFTP   Means that the stdout from the job will be
#                                           stored during job execution on the batch worker
#                                           and then transfered back by GSIFTP at the end.
#                                           If CACHE_EXPORT_DIR (see later) is defined the
#                                           stdout is returned as a file in the local gass cache
#                                           and appeneded to the gass cache on the gatekeeper.
#                                           Otherwise an explict gsiftp is made and the file
#                                           pointed to by REMOTE_STDOUT is overwritten. NB
#                                           this mode of operation could cause problems with
#                                           multi count jobs
#
#                             Otherwise the stdout is assumed to be handled by the batch system.
#
#            REMOTE_STDERR    Similar to REMOTE_STDOUT but curresponding to STDERR instead.
#
#            STDERR_XFER      Similar to STDOUT_XFER but curresponding to STDERR instead.
#
#            CACHE_EXPORT_DIR Either '-' or the location of the cache export/import directory
#                             on the gatekeeper.
#
#            CACHE_TAG        Tag that this job should use for storing files in the cache.
#                             It cache_tag is given as '-', the value from the environment
#                             variable 'GLOBUS_GRAM_JOB_CONTACT' will be used instead.
#
#            GRAM_SCRATCH_DIR Either '-' or the location of a directory in which to make a scratch
#                             directory for this job. An attempt will be made
#                             to make a uniqe directory in this directory on the local machine
#                             after which all references to the original directory in the job's
#                             environment will be changed to the new one. The scratch directory
#                             and all contents will be removed at the end of the job.
#
#            EMERGX509        Method to be used incase the current proxy is invalid or has
#                             expired.
#
#                             NONE        No method to obtain an emergency proxy
#
#                             PBSMSG      Read the batch system's stderr file and
#                                         look for a proxy there. This can be used with
#                                         PBSs qmsg facility to send a 'message' to the
#                                         log file of a running job.
#
#            JOB_EXEC_PATH    The path or URL of the executable to be run. If it is a URL that
#                             URL is lookedup in the local gass cache, if the URL exists the
#                             corresponding file in the cache is used. If this is a path and
#                             CACHE_EXPORT_DIR is being used, and the path exists in the gatekeeper's
#                             gass cache, it will be automaticaly translated to the path of
#                             the executable transfered over to the local gass cache.
#                             Otherwise if JOB_EXEC_PATH is an absolute path and it exists and
#                             is executable on the batch worker the local executable will be directly
#                             started. If the path does not exist on the local machine the path
#                             is taken to be the path to the executable on the gatekeeper.
#                             The executable will be transfered over from the gatekeeper to a
#                             temporary file and run from that.
#             
#            EXEC_ARGS        Space separated list of arguments to be supplied to the
#                             executable.
#
#
# If defined, the CACHE_EXPORT_DIR on the gatekeeper should contain an archive cache_export_dir.tar, which
# in turn contains:
#
#   export.txt                Space separated list of: (tag, url, remote_filename, link_name) entries in
#                             the gatekeeper's gass cache that should be copied over and
#                             entered into the job's local gass cache.
#
#                               The following will not be copied over:
#
#                                 URLs in the gatekeeper's cache that have any tag with
#                                 a file path equal to REMOTE_STDOUT, REMOTE_STDERR,
#                                 REMOTE_X509_NAME or GLOBUS_REMOTE_IO_URL (if defined).
#
#                                 URLs that end in 'gk_cache_export_dir'
#
#   stage_in.txt              List of file names to copy from the gatekeeper to the working
#                             directory of the job. The list is built from the file_stage_in
#                             and file_stage_in_shared directives for the job. Both are treated
#                             in the same way:
#
#                             If the file already exists on the batch worker it is used,
#                             otherwise an attempt is made to copy it from the gatekeeper. If
#                             the path for the file, as interpreded for the local machine, does
#                             not exist, or the directory exists but is not writable, the named
#                             file will be placed directly into the inital current working
#                             directory for the job.
#
#                             After the job finishes the files will be removed, if no other job
#                             appears to be using that file as a 'stage in' file, via a shared
#                             file system for instance.
#
#                             (NB: Therefore stage_in files can be removed by a job even if
#                                  that job never staged them)
#
#   stage_out.txt             List of file names to copy back to the gatekeeper.
#                             The stage out files will be removed from the batch worker once
#                             they have been sent to the gatekeeper, unless they are also
#                             listed as a stage_in file.
#
#   file_cleanup.txt          List of files to remove
#                             from the batch worker once the executable has finished.
#
#   <(ref)import.txt>         This file is created and transfered back to the gatekeeper
#                             when the job exits. It lists the contents of the local gass cache,
#                             with tags matching cache_tag, along with references to <import_refs>
#
#                               The following are never copied back:
#
#                                 URLs matching JOB_EXEC_PATH or JOB_STDIN (in case those are
#                                 non-file URLs)
#
#   <import_refs>             Contents of the local gass cache, to allow for importing back
#                             into the cache on the gatekeeper. These are added to a signle
#                             arhive and transfered back to the gatekeeper in one transfer.
#
#                             On the gatekeeper the contents of each file are appeneded to
#                             files under any existing matching tag+URL there.
#

umask 0077;

my @procs_to_kill;
my @files_to_rm;
my @dirs_to_rm;
my $batch_stderr_fd = fileno(STDERR);

chomp(my $inital_pwd = `pwd`);
chomp(my $my_hostname = `hostname`);
$my_hostname =~ s/\..*//;

my @args = @ARGV;
my_exit(1,"wrong number of command line arguments") if scalar(@args)<14;

my $grami_directory = shift @args;
my $gate_hostname = shift @args;
my $remote_x509_name = shift @args;
my $inital_x509 = shift @args;
my $job_stdin = shift @args;
my $remote_stdout = shift @args;
my $stdout_xfer = shift @args;
my $remote_stderr = shift @args;
my $stderr_xfer = shift @args;
my $cache_export_dir = shift @args;
my $cache_tag = shift @args;
my $gram_scratch_dir = shift @args;
my $emergx509 = shift @args;
my $job_exec_path = shift @args;
my @exec_args = @args;

my_exit(1,"could not find globus-url-copy") unless -x $ENV{"GLOBUS_LOCATION"}."/bin/globus-url-copy";
my_exit(1,"could not find globus-gass-cache") unless $cache_export_dir eq '-' || -x $ENV{"GLOBUS_LOCATION"}."/bin/globus-gass-cache";
my_exit(1,"could not find grid-proxy-info") unless $remote_x509_name eq '-' || -x $ENV{"GLOBUS_LOCATION"}."/bin/grid-proxy-info";

my $ret;

my $my_name = $0;
make_absolute(\$my_name,$inital_pwd);
my $home_dir = (getpwuid($>))[7];

if ( -d "$grami_directory" ) {
  my ($dev_h,$ino_h,$dev_g,$ino_g) = ((CORE::stat($home_dir))[0,1],(CORE::stat($grami_directory))[0,1]);

  if (!(defined $dev_h && defined $ino_h &&
        defined $dev_g && defined $ino_g &&
        $dev_h == $dev_g && $ino_h == $ino_g)) {
    chdir $grami_directory || my_exit(1,"could not change directory to $grami_directory");
  } else {
    chdir make_tmp($inital_pwd,"d");
  }
}

chomp(my $pwd = `pwd`);

my $cache_pgm = "source \$GLOBUS_LOCATION/etc/globus-user-env.sh; globus-gass-cache";

$cache_tag = $ENV{'GLOBUS_GRAM_JOB_CONTACT'} if $cache_tag eq '-';

my_exit(1,"remote_stdout ($remote_stdout) is not a file url") if $remote_stdout ne '-' && make_absolute(\$remote_stdout,$grami_directory) ne 'file';
my_exit(1,"remote_stderr ($remote_stderr) is not a file url") if $remote_stderr ne '-' && make_absolute(\$remote_stderr,$grami_directory) ne 'file';
my_exit(1,"remote_x509_name ($remote_x509_name) is not a file url") if $remote_x509_name ne '-' && make_absolute(\$remote_x509_name,$grami_directory) ne 'file';

#
# Initialy we don't know where we would store our stdout,stderr in the local gass cache
#
my ($url_stdout,$url_stderr)=("","");

#
# Setup inital gass cache location for this job
#
$ENV{"GLOBUS_GASS_CACHE_DEFAULT"} = make_tmp($pwd,"d");

#------------------------
#Deal with proxy
#
if ($remote_x509_name ne '-') {
  my $x509_name = make_tmp("/tmp","f");
  open(PROXY,"> ".$x509_name) || my_exit(1,"could not open local proxy temporary file");
# Have opened PROXY #

  if (lc($inital_x509) eq 'x509self') {
# Take proxy appended to this script #
#
    open(SELF,"< ".$my_name) || my_exit(1,"could not open self to extract initial x509 proxy");
    while(<SELF>) {
      chomp(my $line=$_);
      if ($line =~ /^#X509#/) {
        my $output = $line; $output =~ s/^#X509#//;
        print PROXY $output."\n";
      }
    }
    close(SELF);
  } elsif (lc($inital_x509) =~ m|^x509gpg:|) {
#
# Take proxy transfered over by batch system, but encoded with password found in self
#
    my $local = $inital_x509;
    $local =~ s/^[^:]*://;
    my_exit(1,"bad initial encoded proxy filename") if $local eq '';

    my $tmp = $local;
    make_absolute(\$local,$inital_pwd);
    chmod 0600, $local;
    if ( ! -e $local )
    {
        $local = $tmp;
        make_absolute(\$local,$home_dir);
        chmod 0600, $local;

        my_exit(1,"could not find initial, gpg-ed, proxy") if ( ! -e $local );
    }

    my @data;
    if (!decode_gpg($local,\@data)) {
      my_exit(1,"could not read decoded proxy from gpg");
    }

    foreach my $line (@data) {
      print PROXY $line."\n";
    }

    unlink($local);

  } elsif (lc($inital_x509) =~ m|^x509batch:|)  {
#
# Take proxy transfered over by batch system as an input file #
#
    my $local = $inital_x509;
    $local =~ s/^[^:]*://;
    my_exit(1,"bad initial proxy filename") if $local eq '';

    if ($local ne '-')
    {
        my $tmp = $local;
        make_absolute(\$local,$inital_pwd);
        chmod 0600, $local;
        if (!open(LOCALX509,"< ".$local))
        {
            $local = $tmp;
            make_absolute(\$local,$home_dir);
            chmod 0600, $local;
            open(LOCALX509,"< ".$local) || my_exit(1,"could not open initial proxy transfered by batch system");
        }
    }
    else
    {
        open(LOCALX509,"<&STDIN") || my_exit(1,"could not dup STDIN to read the initial proxy");
    }
    while(<LOCALX509>) {
      chomp(my $line=$_);
      print PROXY $line."\n";
    }
    close(LOCALX509);

    unlink($local) if $local ne '-';

  } elsif (lc($inital_x509) =~ m|^none|)  {
#
# Nothing. Had better have an emergency method.
#
  } else {
    my_exit(1,"unknown method specified for obtaining initial X509 proxy");
  }
  close(PROXY);

#
#Set environment to point at new proxy copy and start the process
#to watch for proxy expiration
#

  foreach my $var_name (keys %ENV) {
    $ENV{$var_name} = $x509_name if $ENV{$var_name} eq $remote_x509_name;
  }

  my $x509_tmp = make_tmp("/tmp","f");

  my $child_proxy = fork();
  my_exit(1,"could not fork child to watch proxy") unless defined $child_proxy;
  do_child_watch_proxy($remote_x509_name,$x509_name,$x509_tmp) unless $child_proxy;
#
# Only parent should reach this point #
#
  push(@procs_to_kill,$child_proxy);

  wait_for_valid_proxy(1200) if lc($emergx509) ne 'none';
}
#
# End of Proxy stuff
#------------------------

#
# Handle cache_export_dir. If cache_export_dir is defined:
#  Populate the job's local gass cache
#  Fetch and set stage_in, stage_out and file_cleanup lists
#  Lookup the URLs of the STDOUT,STDERR in the gatekeeper's cache
#

my ($stage_in_list,$stage_out_list,$file_cleanup_list) = ('-','-','-');
my $local_cache_dir = make_tmp($pwd,"d");

if ($cache_export_dir ne '-') {
  my_exit(1,"location of cache export directory not a file url") if make_absolute(\$cache_export_dir,$grami_directory) ne 'file';

  $ret = get_gk_file($cache_export_dir."/cache_export_dir.tar",$local_cache_dir."/cache_export_dir.tar");

  if ($ret==0 && -e "$local_cache_dir/cache_export_dir.tar") {
    system("cd $local_cache_dir; tar xf cache_export_dir.tar >/dev/null 2>&1");
  } else {
    my_exit(1,"cache_export_dir ($cache_export_dir) on gatekeeper did not contain a cache_export_dir.tar archive");
  }

  my %transfered_url;
  my %ignore_url;

  open(IMPORTLIST,"< ".$local_cache_dir."/export.txt") || my_exit(1,"Could not open export.txt");
  while(<IMPORTLIST>) {
    chomp(my $line=$_);
    my ($tag,$url,$remote_file,$link_name) = split('\s+',$line);
      
    $url_stdout = $url if $remote_stdout ne '-' && $remote_stdout eq $remote_file && $tag eq $cache_tag;
    $url_stderr = $url if $remote_stderr ne '-' && $remote_stderr eq $remote_file && $tag eq $cache_tag;

    if (($remote_stdout ne '-' && $remote_stdout eq $remote_file) ||
        ($remote_stderr ne '-' && $remote_stderr eq $remote_file) ||
        ($remote_x509_name ne '-' && $remote_x509_name eq $remote_file) ||
        ($url =~ m|/gk_cache_export_dir$|) ||
        (defined $ENV{"GLOBUS_REMOTE_IO_URL"} && $ENV{"GLOBUS_REMOTE_IO_URL"} eq $remote_file)) {

      $ignore_url{$url} = 1;
    }

    next if defined $ignore_url{$url};

    system($cache_pgm." -add -t $tag -n $url file:///dev/null >/dev/null 2>&1");

    chomp(my $local = `$cache_pgm -query -t $tag $url`);

    if ($local ne '' && -e "$local") {
      my $rc=0;

      if (! defined $transfered_url{$url}) {
        $rc = res_to_exitval(system("cat $local_cache_dir/".$link_name." > ".$local." 2>/dev/null"));
      }

      if ($rc != 0) {
        system($cache_pgm." -delete -t $tag $url >/dev/null 2>&1");
      } else {
        $transfered_url{$url} = 1;
        foreach my $var_name (keys %ENV) {
          $ENV{$var_name} = $local if $ENV{$var_name} eq $remote_file;
        }
        if ($job_stdin eq $remote_file && $tag eq $cache_tag) {
          $job_stdin = $local;
        }
        if ($job_exec_path eq $remote_file && $tag eq $cache_tag) {
          $job_exec_path = $local;
          chmod 0700, $job_exec_path;
        }
      }
    }
  }
  close(IMPORTLIST);

  file_list_read($local_cache_dir."/stage_in.txt",\$stage_in_list);

  file_list_read($local_cache_dir."/stage_out.txt",\$stage_out_list);

  file_list_read($local_cache_dir."/file_cleanup.txt",\$file_cleanup_list);
}

#
# Handle the GLOBUS_REMOTE_IO_URL lookup file, if it is defined
#
if (defined $ENV{"GLOBUS_REMOTE_IO_URL"}) {
  my ($io_pipe_name,$io_fifo_tmp1,$io_fifo_tmp2);

  $io_pipe_name = make_tmp("/tmp","p");
  $io_fifo_tmp1 = make_tmp("/tmp","f");
  $io_fifo_tmp2 = make_tmp("/tmp","f");

  my $remote_file = $ENV{"GLOBUS_REMOTE_IO_URL"};
  my $child_io = fork();
  my_exit(1,"could not fork a child to watch for reads on the globus_remote_io_url pipe") unless defined $child_io;
  do_child_io_pipe($remote_file,$io_pipe_name,$io_fifo_tmp1,$io_fifo_tmp2) unless $child_io;
#
# Only parent should reach this point #
#
  push(@procs_to_kill,$child_io);

  foreach my $var_name (keys %ENV) {
    $ENV{$var_name} = $io_pipe_name if $ENV{$var_name} eq $remote_file;
  }
}

#
# Try to create a unique scratch directory, in the directory used for $gram_scratch_dir
#
if ($gram_scratch_dir ne '-') {
  my_exit(1,"gram_scratch_dir was not an absolue path") unless $gram_scratch_dir =~ m|^/|;
  my @path = split("/",$gram_scratch_dir);

  if (! -d $gram_scratch_dir) {
    pop(@path);
  }

  my $orig_dir = join("/",@path);
  my_exit(1,"could not get a good path for the temporary directory (tried $orig_dir)") unless -d "$orig_dir" && -w "$orig_dir";

  my $new_scratch = make_tmp($orig_dir,"d");
  foreach my $var (keys %ENV) {
    $ENV{$var} = $new_scratch if $ENV{$var} eq $gram_scratch_dir;
  }
}

#
# Deal with the STAGE_IN files
#
my %staged_in_to_remove;
if ($stage_in_list ne '-') {

  my @stage_in = split(",",$stage_in_list);
  foreach my $file (@stage_in) {
    my $local = $file;
    next if make_absolute(\$local,$pwd) ne 'file';

    # Here we test that the incoming path is somewhere reasonable #

    my @path = split("/",$local);
    my $file = pop(@path);
    my $orig_dir = join("/",@path);

    if ( ! -d "$orig_dir" || (! -e "$local" && ! -w "$orig_dir")) {
      $local = $pwd."/".$file;
      $orig_dir = $pwd;
    }

    my $instance_lock = $local.".lock.$my_hostname.$$";
    my $global_lock = $local.".lock";

    next unless open(LOCK,"> $instance_lock");  # Perhaps the directory is read only #
    close(LOCK);unlink($instance_lock);

#
# Check: For each instance lock check that there exists a corresponding process
#
    opendir(DIR,$orig_dir."/.");
    foreach my $dir_entry (readdir(DIR)) {
      if ($dir_entry =~ m|^$file\.lock\.$my_hostname\.(\d+)$|) {
        my $pid = $1;
        my $exists = kill(0,$pid);
        unlink($orig_dir."/".$dir_entry) unless $exists;
      }
    }
    closedir(DIR);

    my ($got_lock,$attempts) = (0,0);
    while(!$got_lock && $attempts<60) {
      $attempts++;
      if (defined sysopen(LOCK, $global_lock, O_WRONLY | O_EXCL | O_CREAT)) {
        close(LOCK);
        $got_lock++;
        if (! -e "$local") {
          my $remote = $file;
          make_absolute(\$remote,$grami_directory);
          $ret = get_gk_file($remote,$local);
          unlink($local) unless $ret == 0;
        }
        if ( -e "$local") {
          $ret = res_to_exitval(system("ln ".$local." ".$instance_lock));
          if ($ret == 0) {
            $staged_in_to_remove{$local} = $instance_lock.":".$global_lock;

# add the instance lock to file remove list, just incase we exit
# abnormaly before the lock is removed explicitly
            push(@files_to_rm,$instance_lock);

          }
        }
        unlink($global_lock);
      } else {
        if ($attempts > 30) {
          sleep 5;
          check_for_old_global_lock($local,$global_lock,300);
          sleep 5;
        } else {
          sleep 10;
        }
      }
    }
  } # foreach
}

#
# Get original executable #
#
my $local_exec_path;
if ($job_exec_path =~ m|^/| && -x "$job_exec_path" ) {
  $local_exec_path = $job_exec_path;
} else {
  my $full_exec_path = $job_exec_path;
  if (make_absolute(\$full_exec_path,$grami_directory) eq 'file') {
    $local_exec_path = make_tmp($pwd,"f");
    $ret = get_gk_file($full_exec_path,$local_exec_path);
    my_exit(1,"could not fetch the executable ($full_exec_path) from the gatekeeper") if $ret;
  } else {
      my_exit(1,"executable should be in the local gass cache, but no cache export dir used") if $cache_export_dir eq '-';
      chomp(my $local = `$cache_pgm -query -t $cache_tag $full_exec_path`);
      my_exit(1,"could not find the executable in the local gass cache") if $local eq "" || ! -e "$local";
      $local_exec_path = $local;
  }
  chmod 0700, $local_exec_path;
}
#
# Prepare to run original executable
#

my ($tmp_stdin,$tmp_stdout,$tmp_stderr);
if ($job_stdin ne '-') {
    $tmp_stdin = $job_stdin;
    if (make_absolute(\$tmp_stdin,$pwd) eq 'file') {
      if ( ! -r "$tmp_stdin") {
        my $full_job_stdin = $job_stdin;
        make_absolute(\$full_job_stdin,$grami_directory);

        $tmp_stdin = make_tmp($pwd,"f");
        $ret = get_gk_file($full_job_stdin,$tmp_stdin);
        my_exit(1,"could not fetch the stdin from the gatekeeper") if $ret;
      }
    } else {
      my_exit(1,"stdin should be in the local gass cache, but no cache export dir used") if $cache_export_dir eq '-';
      chomp(my $local = `$cache_pgm -query -t $cache_tag $tmp_stdin`);
      my_exit(1,"could not find the stdin in the local gass cache") if $local eq '' || ! -e "$local";
      $tmp_stdin = $local;
    }
    open(OLDIN, "<&STDIN") || my_exit(1,"could not dup stdin to oldin");
    fcntl(OLDIN, F_SETFD, FD_CLOEXEC);
    open(STDIN, "< ".$tmp_stdin) || my_exit(1,"could not open stdin from $tmp_stdin");
}
if ($remote_stdout ne '-' && lc($stdout_xfer) eq "stdoutftp") {
    if ($url_stdout ne "") {
      system($cache_pgm." -add -t $cache_tag -n $url_stdout file:///dev/null >/dev/null 2>&1");
      chomp($tmp_stdout = `$cache_pgm -query -t $cache_tag $url_stdout`);
      my_exit(1,"could not add entry in the local gass cache for stdout") if $tmp_stdout eq "";
    } else {
      $tmp_stdout = make_tmp($pwd,"f");
    }
    open(OLDOUT, ">&STDOUT") || my_exit(1,"could not dup stdout to oldout");
    fcntl(OLDOUT, F_SETFD, FD_CLOEXEC);
    open(STDOUT, "> ".$tmp_stdout) || my_exit(1,"could not open stdout file $tmp_stdout");
    my $tmp_fh = select(STDOUT); $| = 1; select($tmp_fh);    # make stdout unbuffered
}
if ($remote_stderr ne '-' && lc($stderr_xfer) eq "stderrftp") {
    if ($url_stderr ne "") {
      system($cache_pgm." -add -t $cache_tag -n $url_stderr file:///dev/null >/dev/null 2>&1");
      chomp($tmp_stderr = `$cache_pgm -query -t $cache_tag $url_stderr`);
      my_exit(1,"could not add entry in the local gass cache for stderr") if $tmp_stderr eq "";
    } else {
      $tmp_stderr = make_tmp($pwd,"f");
    }
    open(OLDERR, ">&STDERR") || my_exit(1,"could not dup stderr to olderr");
    fcntl(OLDERR, F_SETFD, FD_CLOEXEC);
    open(STDERR, "> ".$tmp_stderr) || my_exit(1,"could not open stderr file $tmp_stderr");
    my $tmp_fh = select(STDERR); $| = 1; select($tmp_fh);    # make stderr unbuffered
}

my $cmd = $local_exec_path." ".join(" ",@exec_args);
my $wrapper = "if [ -x \${LCG_LOCATION:-/opt/lcg}/libexec/jobwrapper ]; then \${LCG_LOCATION:-/opt/lcg}/libexec/jobwrapper $cmd; else $cmd; fi";

umask 0022;
my $old_home = $ENV{"HOME"};
$ENV{"HOME"} = $pwd;
my $cmd_res = system($wrapper);
if (defined $old_home) {
  $ENV{"HOME"} = $old_home;
} else {
  delete $ENV{"HOME"};
}
umask 0077;

if ($job_stdin ne '-') {
  open(STDIN, "<&OLDIN") || my_exit(1,"could not dup stdin from oldin");
  close(OLDIN);
}

if ($remote_stdout ne '-' && lc($stdout_xfer) eq "stdoutftp") {
  open(STDOUT, ">&OLDOUT") || my_exit(1,"could not dup stdout from oldout");
  close(OLDOUT);
  if ($url_stdout eq "") {
    $ret = res_to_exitval(system("source \$GLOBUS_LOCATION/etc/globus-user-env.sh;".
                                 "globus-url-copy file://".$tmp_stdout." gsiftp://".$gate_hostname.$remote_stdout." >/dev/null 2>&1"));
    my_exit(1,"could not return stdout to gatekeeper") if $ret;
  }
}
if ($remote_stderr ne '-' && lc($stderr_xfer) eq "stderrftp") {
  open(STDERR, ">&OLDERR") || my_exit(1,"could not dup stderr from olderr");
  close(OLDERR);
  if ($url_stderr eq "") {
    $ret = res_to_exitval(system("source \$GLOBUS_LOCATION/etc/globus-user-env.sh;".
                                 "globus-url-copy file://".$tmp_stderr." gsiftp://".$gate_hostname.$remote_stderr." >/dev/null 2>&1"));
    my_exit(1,"could not return stderr to gatekeeper") if $ret;
  }
}
#
# Finished with original executable
#

#
# Send back and then remove STAGE_OUT files (unless they were broguht in by STAGE_IN)
#
if ($stage_out_list ne '-') {
  my @stage_out = split(",",$stage_out_list);
  foreach my $file (@stage_out) {
    my $local = $file;
    next if make_absolute(\$local,$pwd) ne 'file';
    if ( -e "$local") {
      my $remote = $file;
      make_absolute(\$remote,$grami_directory);
      $ret = res_to_exitval(system("source \$GLOBUS_LOCATION/etc/globus-user-env.sh;".
                                   "globus-url-copy file://".$local." ".
                                   "gsiftp://".$gate_hostname.$remote." >/dev/null 2>&1"));
      unlink($local) if ($ret == 0 && !defined $staged_in_to_remove{$local});
    }
  }
}

#
# Tidy up STAGE_IN files
#
foreach my $local (keys %staged_in_to_remove) {
  my ($instance_lock,$global_lock) = split(":",$staged_in_to_remove{$local});

  my ($got_lock,$attempts) = (0,0);
  while(!$got_lock && $attempts<60) {
    $attempts++;
    if (defined sysopen(LOCK, $global_lock, O_WRONLY | O_EXCL | O_CREAT)) {
      close(LOCK);
      $got_lock++;
      unlink($instance_lock);
      my $nlinks = (stat($local))[3];
      unlink($local) if defined $nlinks && $nlinks == 1;
      unlink($global_lock);
    } else {
      if ($attempts > 30) {
        sleep 5;
        check_for_old_global_lock($local,$global_lock,300);
        sleep 5;
      } else {
        sleep 10;
      }
    }
  }
}

#
# Export contents (with tag $cache_tag) of local gass cache back to gatekeeper,
#   except job executable or stdin
#
if ($cache_export_dir ne '-') {
  my $ref = $my_hostname.".".$$;
  my $master_import_index = $ref.".import.txt";

  my %cache_url_tags;
  my $cache_url;

  open(CACHEPRG,$cache_pgm." -list -t $cache_tag |") || my_exit(1,"could not open pipe from cache prog during cache export (1)");
  while(<CACHEPRG>) {
    chomp(my $line=$_);
    if ($line =~ /^\s*URL:\s*(\S+)/) {
      $cache_url = $1;
    }
    if (defined $cache_url && $line =~ /^\s*Tag:\s*(\S+)/) {
      my $tag = $1;
      if (defined $cache_url_tags{$cache_url}) {
        $cache_url_tags{$cache_url} .= " ".$tag;
      } else {
        $cache_url_tags{$cache_url} = $tag;
      }
      $cache_url = undef;
    }
  }
  close(CACHEPRG);

  my $n_urls=0;

  open(CACHEEXPORT,"> ".$local_cache_dir."/".$master_import_index) || my_exit(1,"could not write to temporary file to write contents of local gass cache");
  foreach $cache_url (keys %cache_url_tags) {

    my @tags =  split('\s+',$cache_url_tags{$cache_url});
    my $has_my_tag=0;
    foreach my $tag (@tags) {
      $has_my_tag++ if $tag eq $cache_tag;
    }

    my $full_exec_path = $job_exec_path;
    if (make_absolute(\$full_exec_path,$grami_directory) ne 'file') {
      next if $has_my_tag && $cache_url eq $full_exec_path;
    }
    my $full_stdin_path = $job_stdin;
    if (make_absolute(\$full_stdin_path,$pwd) ne 'file') {
      next if $has_my_tag && $cache_url eq $full_stdin_path;
    }

    open(CACHEPRG,$cache_pgm." -query -t $tags[0] $cache_url |") || my_exit(1,"could not open pipe from cache prog during cache export (2)");
    my $file;
    while(<CACHEPRG>) {
      chomp($file = $_);
    }
    close(CACHEPRG);
    next if ($file eq "" || ! -e "$file");

    $n_urls++;
    my $file_ref = $ref.".".$n_urls;
    symlink($file,$local_cache_dir."/".$file_ref);

    foreach my $tag (@tags) {
      print CACHEEXPORT $tag." ".$cache_url." ".$file_ref."\n";
    }

  }
  close(CACHEEXPORT);

  $ret = res_to_exitval(system("cd $local_cache_dir; tar chf $master_import_index.tar $master_import_index >/dev/null 2>&1"));
  my_exit(1,"could not create initial master_import_index.tar") if $ret != 0;

  for(my $i=1;$i<=$n_urls;$i++) {

    $ret = res_to_exitval(system("cd $local_cache_dir; tar rhf $master_import_index.tar $ref.$i >/dev/null 2>&1"));
    my_exit(1,"could not add to local gass cache import.txt.tar file") if $ret != 0;

  }

  $ret = res_to_exitval(system("source \$GLOBUS_LOCATION/etc/globus-user-env.sh;".
                               "globus-url-copy file://".$local_cache_dir."/".$master_import_index.".tar ".
                               "gsiftp://".$gate_hostname.$cache_export_dir."/".$master_import_index.".tar >/dev/null 2>&1"));
  my_exit(1,"could not export the local gass cache import.txt.tar file back to the gatekeeper") if $ret != 0;
}

#
# Remove files in the FILE_CLEANUP list, unless they are files
# brought in for stage_in. Don't remember those as they should
# be removed when the last job that needs them finishes.
#
if ($file_cleanup_list ne '-') {
  my @file_cleanup = split(",",$file_cleanup_list);
  foreach my $file (@file_cleanup) {
    my $local=$file;
    next if make_absolute(\$local,$pwd) ne 'file';
    unlink($local) unless defined $staged_in_to_remove{$local};
  }
}

my_exit(res_to_exitval($cmd_res),"");
exit 0;

#############################

sub do_child_io_pipe {
  my ($file,$fifoname,$tmp1,$tmp2) = @_;

  my ($rin,$win,$ein);
  my ($rout,$wout,$eout);
  $SIG{PIPE} = 'IGNORE';
  my $last_time = -1;
  my $ret;
  my $nfound;

  while(1) {
    open(FIFOWRITE,"> ".$fifoname) || exit(1);
    select(FIFOWRITE); $| = 1;

    $rin = $win = $ein = '';
    vec($win,fileno(FIFOWRITE),1) = 1;

    do {
      $nfound = select($rout=$rin, $wout=$win, $eout=$ein, undef);
    } while($nfound<0 && $! =~ /Interrupted/);

    if ($last_time==-1 || time()-$last_time>30) {
      $ret = get_gk_file($file,$tmp2);
      if ($ret == 0) {
        $last_time = time();
        system("cp $tmp2 $tmp1");
      }
    }
    if (-e $tmp1) {
      open(SRC,"< ".$tmp1) || exit(1);
      while(<SRC>) {
        chomp(my $line = $_);
        print $line."\n";
      }
      close(SRC);
    } else {
      print "\n";
    }
    close(FIFOWRITE);
    sleep 2;
  }
  exit 0;
}

sub res_to_exitval {
  my ($result) = @_;

  my ($exit_value,$signal_num,$dumped_core)  = ( $result >> 8,
                                                 $result & 0x7F,
                                                 ($result & 0x80) ? 1 : 0 );
  return $exit_value;
}

sub wait_for_valid_proxy {
  my ($wait_time) = @_;

  my $ttl = -1;
  my $start_time = time();
  my $min_lifetime = 660;

  while(time() - $start_time < $wait_time) {
    if (open(GPI,"source \$GLOBUS_LOCATION/etc/globus-user-env.sh;".
                 "grid-proxy-info 2>&1 |")) {
      my ($hr,$min,$sec);
      my $good=0;
      while(<GPI>) {
        chomp(my $line=$_);
        if ($line =~ /^timeleft\s+:\s+(\d+):(\d+):(\d+)/) {
          ($hr,$min,$sec) = ($1,$2,$3);
          $good++;
        }
      }
      if ($good == 1) {
        $ttl = $hr*3600 + $min*60 + $sec;
      }
      close(GPI);
    }
    last if $ttl>$min_lifetime;
    sleep 10;
  }
}

sub pbsmsg {
  my ($tmp) = @_;
  my $ret = 0;

  my $std_err_filename;
  if (open(LSOF,"/usr/sbin/lsof -a -p ".$$." -d ".$batch_stderr_fd." |")) {
    while(<LSOF>) {
      chomp(my $line=$_);
      if ($line =~ /^(\S+)\s+(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\d+)\s+(\d+)\s+(\S+)/) {
        my ($command,$pid,$user,$fd,$type,$device,$size,$node,$name) = ($1,$2,$3,$4,$5,$6,$7,$8,$9);
        $std_err_filename = $name;
      }
    }
    close(LSOF);
  }
  if (! defined $std_err_filename) {
    if (defined $ENV{"TMPDIR"} && defined $ENV{"PBS_JOBID"}) {
      my @path = split('/',$ENV{"TMPDIR"});
      pop(@path);
      my $hashed_jid = substr($ENV{"PBS_JOBID"},0,11);
      $std_err_filename = join('/',@path)."/spool/".$hashed_jid.".ER";
    }
  }
  if (defined $std_err_filename && -f $std_err_filename && -r $std_err_filename) {
    if (open(SE,"< ".$std_err_filename)) {
      if (open(PROXY,"> ".$tmp)) {
        my $started=0;
        my $whole_proxy=0;
        my $start_pos;
        while(<SE>) {
          chomp(my $line=$_);
          if ($line =~ /^#X509GPG#START$/) {
            $whole_proxy = 0;
            $started = 1;
            $start_pos = tell(SE);
          }
          if ($started && $line =~ /^#X509GPG#FINISH$/) {
            $whole_proxy=1;
            $started=0;
          }
        }
        if ($whole_proxy) {
          seek SE,$start_pos,0;
          while(<SE>) {
            chomp(my $line=$_);
            last if ($line =~ /^#X509GPG#FINISH$/);

            my $output = $line; $output =~ s/^#X509GPG#//;
            print PROXY $output."\n";
          }
          $ret = 1;
        }
        close(PROXY);
      }
      close(SE);
    }
  }

  return $ret;
}

sub do_child_watch_proxy {
  my ($remote_file,$local_name,$tmp)=@_;

  my $ret;
  my $ttl = -1;
  my $emerg_fail = 0;

  while(1) {
    if ($ttl != -1) {
      $ret = get_gk_file($remote_file,$tmp);
      system("cp $tmp $local_name >/dev/null 2>&1") if $ret==0;
    }

    $ttl = -1;
    $ret = res_to_exitval(system("source \$GLOBUS_LOCATION/etc/globus-user-env.sh;".
                                 "grid-proxy-info >$tmp 2>&1"));
    if ($ret == 0) {
      my $good = 0;
      my ($hr,$min,$sec);
      open(RC,"< ".$tmp) || exit(1);
      while(<RC>){
        chomp(my $line=$_);
        if ($line =~ /^timeleft\s+:\s+(\d+):(\d+):(\d+)/) {
          ($hr,$min,$sec) = ($1,$2,$3);
          $good++;
        }
      }
      if ($good == 1) {
        $ttl = $hr*3600 + $min*60 + $sec;
      }
    }

    my $sleep_time;

    if ($ttl>0) {
        $sleep_time = int((0.8+0.1*rand())*($ttl-660));
        $sleep_time = 60 if $sleep_time<60;
    } else {
        $sleep_time = 600;
    }

    if ($ttl<660) {
      if (lc($emergx509) eq 'pbsmsg') {
        my $ret = pbsmsg($tmp);
        my @data;
        if ($ret && decode_gpg($tmp,\@data)) {
          open(X509,"> $local_name");
          foreach my $line (@data) {
            print X509 $line."\n";
          }
          close(X509);
          $ttl = -1;
          $sleep_time = 0;
        } else {
          $emerg_fail++;
          $sleep_time = 10 if ($emerg_fail <= 120);
        }
      }
    }

    sleep $sleep_time;
  }
  exit 0;
}

sub my_exit {
  my ($ret_val,$errmsg) = @_;

  chdir $inital_pwd;

  kill(9,@procs_to_kill) if scalar(@procs_to_kill) > 0;
  unlink(@files_to_rm) if scalar(@files_to_rm) > 0;
  foreach my $dir (@dirs_to_rm) {
    system("rm -rf $dir");
  }

  if ($errmsg ne "") {
    my $line = "submit-helper script running on host $my_hostname gave error: ".$errmsg;
    print STDERR $line."\n";
  }

  exit($ret_val);
}

sub make_tmp {
  my ($base_dir,$file_type) = @_;

  my $file;
  my ($try,$good) = (0,0);

  while(!$good && $try<=1000) {
    my $rint = $try;
    $try++;

    $file = $base_dir."/globus-tmp.$my_hostname.$$.$rint";

    if (! -e "$file") {
      if (lc($file_type) eq 'f') {
        next unless defined sysopen(NEW, $file, O_WRONLY | O_EXCL | O_CREAT);
        close(NEW);
        $good++;
      } elsif (lc($file_type) eq 'p') {
        my $ret = res_to_exitval(system("mkfifo ".$file." >/dev/null 2>&1"));
        next if $ret != 0;
        $good++;
      } elsif (lc($file_type) eq 'd') {
        my $res = mkdir($file, 0700);
        next if $res==0 || -l $file || !-d $file || !-o $file;
        $good++;
      } else {
        my_exit(1,"Unknown file type ".$file_type." in make_tmp");
      }
    }
  }

  my_exit(1,"could not make a temporary file/pipe/directory in $base_dir") unless $good;
  if (lc($file_type) eq 'd') {
    push(@dirs_to_rm,$file);
  } else {
    push(@files_to_rm,$file);
  }

  local(*DIR);
  if (opendir(DIR,$base_dir)) {
    foreach my $entry (readdir(DIR)) {
      if ($entry =~ /^globus-tmp\.(\S+)\.(\d+)\.(\d+)$/) {
        my ($host,$pid,$seq) = ($1,$2,$3);
        $host =~ s/\..*//;
        if ( -o $base_dir."/".$entry && $host eq $my_hostname && kill(0,$pid) != 1) {
          system("rm -rf $base_dir/$entry >/dev/null 2>&1");
        }
      }
    }
    close(DIR);
  }

  return $file;
}

sub make_absolute {
  my ($name_ref,$dir) = @_;

  my $url;
  if ($$name_ref =~ m|^(\S+)://|) {
    $url = $1;
  } else {
    $url = "file";
  }
  
  return $url if $url ne 'file';

  $$name_ref =~ s|^file://||;
  $$name_ref = $dir."/".$$name_ref unless $$name_ref =~ m|^/|;

  return $url;
}

sub check_for_old_global_lock {
  my ($local_file,$global_lock,$min_age) = @_;

# Is the global lock stale?
# Try to test this by seeing if local_file doesn't exist or exists but
# has not been modified recently, and that the global lock file is indeed old.

   my $mtime_file = (stat($local_file))[9];
   my $mtime_lock = (stat($global_lock))[9];
   if ((!defined $mtime_file || time()-$mtime_file >= $min_age) &&
        defined $mtime_lock && time()-$mtime_lock >= $min_age) {
# Looks stale. Run the risk and remove it.
     unlink($global_lock);
   }
}

sub file_list_read {
  my ($import_list_fn,$list_ref) = @_;

  $$list_ref = '-';

  my @list;
  open(IMPORTLIST,"< ".$import_list_fn) || return;
  while(<IMPORTLIST>) {
    chomp(my $line=$_);
    push(@list,$line);
  }
  close(IMPORTLIST);

  $$list_ref = join(",",@list) if scalar(@list)>0;
}

sub get_gk_file {
  my ($remote_file,$local_file) = @_;
  my $ret = res_to_exitval(system("source \$GLOBUS_LOCATION/etc/globus-user-env.sh;".
                                  "globus-url-copy gsiftp://".$gate_hostname.$remote_file." ".
                                  "file://".$local_file." >/dev/null 2>&1"));
  return $ret;
}

sub decode_gpg
{
  my ($gpg_file,$data_ref) = @_;

  my $passphrase;
  open(SELF,"< ".$my_name) || my_exit(1,"could not open self to extract gpg proxy passphrase");
  while(<SELF>) {
    chomp(my $line=$_);
      if ($line =~ /^#X509GPG#/) {
        $passphrase = $line; $passphrase =~ s/^#X509GPG#//;
    }
  }
  close(SELF);

  if (!defined $passphrase) {
    if (open(DATA,"cat $gpg_file 2>/dev/null |")) {
      @$data_ref = ();
      while(<DATA>) {
        chomp(my $line = $_);  
        push(@$data_ref,$line);
      }
      close(DATA);
      return 1;
    }
    return 0;
  }

  pipe RH,WH;
  fcntl(RH, F_SETFD, 0);

  print WH $passphrase."\n";
  close(WH);

  my $got_data = 0;
  if (open(DATA,"cat $gpg_file | gpg --decrypt --batch --no-tty --homedir ".$ENV{"GLOBUS_GASS_CACHE_DEFAULT"}." ".
                                    "--no-options --passphrase-fd ".fileno(RH)." 2>/dev/null |")) {
    $got_data++;
    @$data_ref = ();
    while(<DATA>) {
      chomp(my $line = $_);
      push(@$data_ref,$line);
    }
    close(DATA);
  }

  close(RH);

  if ($got_data != 0) {
    return 1;
  } else {
    return 0;
  }
}
